import logging
import sys
from typing import List, Union

from ibm_watson_machine_learning.preprocessing import DataJoinGraph
from ibm_watson_machine_learning.utils.autoai.enums import (
    TShirtSize, ClassificationAlgorithms, RegressionAlgorithms, PredictionType, Metrics, DataConnectionTypes,
    PipelineTypes, PositiveLabelClass)
from ibm_watson_machine_learning.utils.autoai.errors import LocalInstanceButRemoteParameter, MissingPositiveLabel
from ibm_watson_machine_learning.utils.autoai.utils import is_ipython, check_dependencies_versions
from ibm_watson_machine_learning.workspace import WorkSpace
from .engines import WMLEngine
from .optimizers import LocalAutoPipelines, RemoteAutoPipelines
from .runs import AutoPipelinesRuns, LocalAutoPipelinesRuns
from ..base_experiment.base_experiment import BaseExperiment

__all__ = [
    "AutoAI"
]
# note: disable logging if we are in Jupyter Notebook (lale logs appear in wrong cell)
if is_ipython:
    logging.disable(sys.maxsize)

# note: checking installed versions of scikit, xgboost and lightgbm
check_dependencies_versions()


class AutoAI(BaseExperiment):
    """
    AutoAI class for pipeline models optimization automation.

    Parameters
    ----------
    wml_credentials: dictionary, required
        Credentials to Watson Machine Learning instance.

    project_id: str, optional
        ID of the Watson Studio project.

    space_id: str, optional
        ID of the Watson Studio Space.

    Example
    -------
    >>> from ibm_watson_machine_learning.experiment import AutoAI
    >>> # Remote version of AutoAI
    >>> experiment = AutoAI(
    >>>        wml_credentials={
    >>>              "apikey": "...",
    >>>              "iam_apikey_description": "...",
    >>>              "iam_apikey_name": "...",
    >>>              "iam_role_crn": "...",
    >>>              "iam_serviceid_crn": "...",
    >>>              "instance_id": "...",
    >>>              "url": "https://us-south.ml.cloud.ibm.com"
    >>>            },
    >>>         project_id="...",
    >>>         space_id="...")
    >>>
    >>> # Local version of AutoAI
    >>> experiment = AutoAI()
    """
    # note: initialization of AutoAI enums as class properties
    ClassificationAlgorithms = ClassificationAlgorithms
    RegressionAlgorithms = RegressionAlgorithms
    TShirtSize = TShirtSize
    PredictionType = PredictionType
    Metrics = Metrics
    DataConnectionTypes = DataConnectionTypes
    PipelineTypes = PipelineTypes

    def __init__(self,
                 wml_credentials: Union[dict, 'WorkSpace'] = None,
                 project_id: str = None,
                 space_id: str = None) -> None:
        # note: as workspace is not clear enough to understand, there is a possibility to use pure
        # wml credentials with project and space IDs, but in addition we
        # leave a possibility to use a previous WorkSpace implementation, it could be passed as a first argument
        if wml_credentials is None:
            self._workspace = None
            self.runs = LocalAutoPipelinesRuns()

        else:
            if isinstance(wml_credentials, WorkSpace):
                self._workspace = wml_credentials
            else:
                self._workspace = WorkSpace(wml_credentials=wml_credentials.copy(),
                                            project_id=project_id,
                                            space_id=space_id)

            self.project_id = self._workspace.project_id
            self.space_id = self._workspace.space_id
            self.runs = AutoPipelinesRuns(engine=WMLEngine(self._workspace))
            self.runs._workspace = self._workspace
        # --- end note

    def runs(self, *, filter: str) -> Union['AutoPipelinesRuns', 'LocalAutoPipelinesRuns']:
        """
        Get the historical runs but with WML Pipeline name filter (for remote scenario).
        Get the historical runs but with experiment name filter (for local scenario).

        Parameters
        ----------
        filter: str, required
            WML Pipeline name to filter the historical runs.
            or experiment name to filter the local historical runs.

        Returns
        -------
        AutoPipelinesRuns or LocalAutoPipelinesRuns

        Example
        -------
        >>> from ibm_watson_machine_learning.experiment import AutoAI
        >>> experiment = AutoAI(...)
        >>>
        >>> experiment.runs(filter='Test').list()
        """

        if self._workspace is None:
            return LocalAutoPipelinesRuns(filter=filter)

        else:
            return AutoPipelinesRuns(engine=WMLEngine(self._workspace.wml_client), filter=filter)

    def optimizer(self,
                  name: str,
                  *,
                  prediction_type: 'PredictionType',
                  prediction_column: str,
                  scoring: 'Metrics',
                  desc: str = None,
                  test_size: float = 0.1,
                  max_number_of_estimators: int = 1,
                  train_sample_rows_test_size: float = None,
                  daub_include_only_estimators: List[Union['ClassificationAlgorithms', 'RegressionAlgorithms']] = None,
                  data_join_graph: 'DataJoinGraph' = None,
                  csv_separator: Union[List[str], str] = ',',
                  excel_sheet: Union[str, int] = 0,
                  positive_label: str = None,
                  data_join_only: bool = False,
                  **kwargs) -> Union['RemoteAutoPipelines', 'LocalAutoPipelines']:
        """
        Initialize an AutoAi optimizer.

        Parameters
        ----------
        name: str, required
            Name for the AutoPipelines

        prediction_type: PredictionType, required
            Type of the prediction.

        prediction_column: str, required
            name of the target/label column

        scoring: Metrics, required
            Type of the metric to optimize with.

        desc: str, optional
            Description

        test_size: float, optional
            Percentage of the entire dataset to leave as a holdout. Default 0.1

        max_number_of_estimators: int, optional
            Maximum number (top-K ranked by DAUB model selection) of the selected algorithm, or estimator types,
            for example LGBMClassifierEstimator, XGBoostClassifierEstimator, or LogisticRegressionEstimator
            to use in pipeline composition.  The default is 1, where only the highest ranked by model
            selection algorithm type is used. (min 1, max 4)

        train_sample_rows_test_size: float, optional
            Training data sampling percentage

        daub_include_only_estimators: List[Union['ClassificationAlgorithms', 'RegressionAlgorithms']], optional
            List of estimators to include in computation process.
            See: AutoAI.ClassificationAlgorithms or AutoAI.RegressionAlgorithms

        csv_separator: Union[List[str], str], optional
            The separator, or list of separators to try for separating
            columns in a CSV file.  Not used if the file_name is not a CSV file.
            Default is ','.

        excel_sheet: Union[str, int], optional
            Name or number of the excel sheet to use. Only use when xlsx file is an input.
            Default is 0.

        positive_label: str, optional
            The positive class to report when binary classification.
            When multiclass or regression, this will be ignored.

        t_shirt_size: TShirtSize, optional
            The size of the remote AutoAI POD instance (computing resources). Only applicable to a remote scenario.
            See: AutoAI.TShirtSize

        data_join_graph: DataJoinGraph, optional
            A graph object with definition of join structure for multiple input data sources.
            Data preprocess step for multiple files.

        data_join_only: bool, optional
            If True only preprocessing will be executed.

        Returns
        -------
        RemoteAutoPipelines or LocalAutoPipelines, depends on how you initialize the AutoAI object.

        Example
        -------
        >>> from ibm_watson_machine_learning.experiment import AutoAI
        >>> experiment = AutoAI(...)
        >>>
        >>> optimizer = experiment.optimizer(
        >>>        name="name of the optimizer.",
        >>>        prediction_type=AutoAI.PredictionType.BINARY,
        >>>        prediction_column="y",
        >>>        scoring=AutoAI.Metrics.ROC_AUC_SCORE,
        >>>        desc="Some description.",
        >>>        test_size=0.1,
        >>>        max_num_daub_ensembles=1,
        >>>        train_sample_rows_test_size=1,
        >>>        daub_include_only_estimators=[AutoAI.ClassificationAlgorithms.LGBM, AutoAI.ClassificationAlgorithms.XGB],
        >>>        t_shirt_size=AutoAI.TShirtSize.L
        >>>    )
        >>>
        >>> optimizer = experiment.optimizer(
        >>>        name="name of the optimizer.",
        >>>        prediction_type=AutoAI.PredictionType.MULTICLASS,
        >>>        prediction_column="y",
        >>>        scoring=AutoAI.Metrics.ROC_AUC_SCORE,
        >>>        desc="Some description.",
        >>>    )
        """
        if data_join_graph:
            data_join_graph.problem_type = prediction_type
            data_join_graph.target_column = prediction_column

        if (prediction_type == PredictionType.BINARY and scoring in vars(PositiveLabelClass).values()
                and positive_label is None):
            raise MissingPositiveLabel(scoring, reason=f"\"{scoring}\" needs a \"positive_label\" "
                                                       f"parameter to be defined when used with binary classification.")

        if self._workspace is None and kwargs.get('t_shirt_size'):
            raise LocalInstanceButRemoteParameter(
                "t_shirt_size",
                reason="During LocalOptimizer initialization, \"t_shirt_size\" parameter was provided. "
                       "\"t_shirt_size\" parameter is only applicable to the RemoteOptimizer instance."
            )
        elif self._workspace is None:
            return LocalAutoPipelines(
                name=name,
                prediction_type='classification' if prediction_type in ['binary', 'multiclass'] else prediction_type,
                prediction_column=prediction_column,
                scoring=scoring,
                desc=desc,
                test_size=test_size,
                max_num_daub_ensembles=max_number_of_estimators,
                train_sample_rows_test_size=train_sample_rows_test_size,
                daub_include_only_estimators=daub_include_only_estimators,
                positive_label=positive_label,
                _force_local_scenario=kwargs.get('_force_local_scenario', False)
            )

        else:
            optimizer = RemoteAutoPipelines(
                name=name,
                prediction_type=prediction_type,
                prediction_column=prediction_column,
                scoring=scoring,
                desc=desc,
                test_size=test_size,
                max_num_daub_ensembles=max_number_of_estimators,
                t_shirt_size=self._workspace.restrict_pod_size(t_shirt_size=kwargs.get(
                    't_shirt_size', TShirtSize.M if self._workspace.wml_client.ICP else TShirtSize.L)
                ),
                train_sample_rows_test_size=train_sample_rows_test_size,
                daub_include_only_estimators=daub_include_only_estimators,
                data_join_graph=data_join_graph,
                csv_separator=csv_separator,
                excel_sheet=excel_sheet,
                positive_label=positive_label,
                data_join_only=data_join_only,
                engine=WMLEngine(self._workspace),
                notebooks=kwargs.get('notebooks', True),
                autoai_pod_version=kwargs.get('autoai_pod_version', None),
                obm_pod_version=kwargs.get('obm_pod_version', '2.0.6')
            )
            optimizer._workspace = self._workspace
            return optimizer



import unittest
import io
import sys
from ibm_watson_machine_learning.utils.log_util import get_logger
from preparation_and_cleaning import *


class TestWMLClientWithTensorflow(unittest.TestCase):
    deployment_uid = None
    model_uid = None
    scoring_url = None
    logger = get_logger(__name__)

    @classmethod
    def setUpClass(self):
        TestWMLClientWithTensorflow.logger.info("Service Instance: setting up credentials")
        self.wml_credentials = get_wml_credentials()
        self.client = get_client()
        self.model_path = os.path.join(os.getcwd(), 'artifacts', 'tf-saved_model.tar.gz')

    def test_01_service_instance_details(self):
        TestWMLClientWithTensorflow.logger.info("Check client ...")
        self.assertTrue(self.client.__class__.__name__ == 'APIClient')

        TestWMLClientWithTensorflow.logger.info("Getting instance details ...")
        details = self.client.service_instance.get_details()
        TestWMLClientWithTensorflow.logger.debug(details)

        self.assertTrue("published_models" in str(details))
        self.assertEqual(type(details), dict)

    def test_02_publish_local_model_in_repository(self):
        TestWMLClientWithTensorflow.logger.info("Saving trained model in repo ...")
        TestWMLClientWithTensorflow.logger.debug(self.model_path)

        self.client.repository.ModelMetaNames.show()

        model_meta_props = {self.client.repository.ModelMetaNames.NAME: "LOCALLY created agaricus prediction model",
                       self.client.repository.ModelMetaNames.TYPE: "tensorflow_1.15",
                       self.client.repository.ModelMetaNames.RUNTIME_UID: "tensorflow_1.15-py3.6"
                            }
        published_model_details = self.client.repository.store_model(model=self.model_path, meta_props=model_meta_props)
        TestWMLClientWithTensorflow.model_uid = self.client.repository.get_model_uid(published_model_details)
        TestWMLClientWithTensorflow.logger.info("Published model ID:" + str(TestWMLClientWithTensorflow.model_uid))
        self.assertIsNotNone(TestWMLClientWithTensorflow.model_uid)

    def test_04_load_model(self):
        TestWMLClientWithTensorflow.logger.info("Load model from repository: {}".format(TestWMLClientWithTensorflow.model_uid))
        self.tf_model = self.client.repository.load(TestWMLClientWithTensorflow.model_uid)
        TestWMLClientWithTensorflow.logger.debug("TF model type: {}".format(type(self.tf_model)))
        self.assertTrue(self.tf_model)

    def test_05_create_deployment(self):
        TestWMLClientWithTensorflow.logger.info("Create deployment")
        deployment_details = self.client.deployments.create(artifact_uid=TestWMLClientWithTensorflow.model_uid, meta_props={self.client.deployments.ConfigurationMetaNames.NAME: "Test deployment",self.client.deployments.ConfigurationMetaNames.ONLINE:{}})

        TestWMLClientWithTensorflow.deployment_uid = self.client.deployments.get_uid(deployment_details)
        TestWMLClientWithTensorflow.scoring_url = self.client.deployments.get_scoring_href(deployment_details)
        self.assertTrue('online' in str(deployment_details))

    def test_06_scoring(self):
        TestWMLClientWithTensorflow.logger.info("Score model")
        scoring_data = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                                 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                                 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                                 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                                 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                                 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                                 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                                 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                                 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                                 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                                 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                                 0, 0, 0, 0, 0, 0, 0, 0, 0, 3, 18, 18, 18,
                                 126, 136, 175, 26, 166, 255, 247, 127, 0, 0, 0, 0, 0,
                                 0, 0, 0, 0, 0, 0, 0, 30, 36, 94, 154, 170, 253,
                                 253, 253, 253, 253, 225, 172, 253, 242, 195, 64, 0, 0, 0,
                                 0, 0, 0, 0, 0, 0, 0, 0, 49, 238, 253, 253, 253,
                                 253, 253, 253, 253, 253, 251, 93, 82, 82, 56, 39, 0, 0,
                                 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 18, 219, 253,
                                 253, 253, 253, 253, 198, 182, 247, 241, 0, 0, 0, 0, 0,
                                 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                                 80, 156, 107, 253, 253, 205, 11, 0, 43, 154, 0, 0, 0,
                                 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                                 0, 0, 0, 14, 1, 154, 253, 90, 0, 0, 0, 0, 0,
                                 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                                 0, 0, 0, 0, 0, 0, 0, 139, 253, 190, 2, 0, 0,
                                 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                                 0, 0, 0, 0, 0, 0, 0, 0, 0, 11, 190, 253, 70,
                                 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                                 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 35,
                                 241, 225, 160, 108, 1, 0, 0, 0, 0, 0, 0, 0, 0,
                                 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                                 0, 0, 81, 240, 253, 253, 119, 25, 0, 0, 0, 0, 0,
                                 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                                 0, 0, 0, 0, 0, 45, 186, 253, 253, 150, 27, 0, 0,
                                 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                                 0, 0, 0, 0, 0, 0, 0, 0, 16, 93, 252, 253, 187,
                                 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                                 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 249,
                                 253, 249, 64, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                                 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 46, 130,
                                 183, 253, 253, 207, 2, 0, 0, 0, 0, 0, 0, 0, 0,
                                 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 39, 148,
                                 229, 253, 253, 253, 250, 182, 0, 0, 0, 0, 0, 0, 0,
                                 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 24, 114,
                                 221, 253, 253, 253, 253, 201, 78, 0, 0, 0, 0, 0, 0,
                                 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 23, 66,
                                 213, 253, 253, 253, 253, 198, 81, 2, 0, 0, 0, 0, 0,
                                 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 18, 171,
                                 219, 253, 253, 253, 253, 195, 80, 9, 0, 0, 0, 0, 0,
                                 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 55, 172,
                                 226, 253, 253, 253, 253, 244, 133, 11, 0, 0, 0, 0, 0,
                                 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                                 136, 253, 253, 253, 212, 135, 132, 16, 0, 0, 0, 0, 0,
                                 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                                 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                                 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                                 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                                 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                                 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                                 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                                 0, 0, 0, 0]

        scoring_payload = {
            self.client.deployments.ScoringMetaNames.INPUT_DATA : [
                {
                    'values': [scoring_data, scoring_data]
                }
            ]
        }
        scores = self.client.deployments.score(TestWMLClientWithTensorflow.deployment_uid, meta_props=scoring_payload)
        self.assertIsNotNone(scores)

    def test_07_list_models(self):
        TestWMLClientWithTensorflow.logger.info("List models")
        stdout_ = sys.stdout
        stream = io.StringIO()
        sys.stdout = stream
        self.client.repository.list()
        sys.stdout = stdout_
        TestWMLClientWithTensorflow.logger.info(stream.getvalue())
        self.assertTrue('GUID' in stream.getvalue())

    def test_08_list_deployments(self):
        TestWMLClientWithTensorflow.logger.info("List deployments")
        stdout_ = sys.stdout
        stream = io.StringIO()
        sys.stdout = stream
        self.client.deployments.list()
        sys.stdout = stdout_
        TestWMLClientWithTensorflow.logger.info(stream.getvalue())
        self.assertTrue('GUID' in stream.getvalue())

    def test_09_delete_deployment(self):
        TestWMLClientWithTensorflow.logger.info("Delete deployment")
        self.client.deployments.delete(TestWMLClientWithTensorflow.deployment_uid)

    def test_10_delete_model(self):
        TestWMLClientWithTensorflow.logger.info("Delete model")
        self.client.repository.delete(TestWMLClientWithTensorflow.model_uid)


if __name__ == '__main__':
    unittest.main()



from ibm_watson_machine_learning.experiment.autoai.optimizers.local_auto_pipelines import LocalAutoPipelines
from ibm_watson_machine_learning.experiment.autoai.optimizers.remote_auto_pipelines import RemoteAutoPipelines



import unittest
import time
import io
import sys
from ibm_watson_machine_learning.utils.log_util import get_logger
from ibm_watson_machine_learning.experiments import Experiments
from preparation_and_cleaning import *


class TestWMLClientWithKeras(unittest.TestCase):
    deployment_uid = None
    definition_1_uid = None
    definition_2_uid = None
    definition_1_url = None
    definition_2_url = None
    experiment_uid = None
    experiment_run_uid = None
    trained_model_uid = None
    train_run_id = None
    scoring_url = None
    logger = get_logger(__name__)

    @classmethod
    def setUpClass(self):
        TestWMLClientWithKeras.logger.info("Service Instance: setting up credentials")
        self.wml_credentials = get_wml_credentials()
        self.client = get_client()
        self.cos_resource = get_cos_resource()
        self.bucket_names = prepare_cos(self.cos_resource, data_code=BOSTON)

    @classmethod
    def tearDownClass(self):
        clean_cos(self.cos_resource, self.bucket_names)

    def test_01_service_instance_details(self):
        TestWMLClientWithKeras.logger.info("Check client ...")
        self.assertTrue(self.client.__class__.__name__ == 'APIClient')

        TestWMLClientWithKeras.logger.info("Getting instance details ...")
        details = self.client.service_instance.get_details()
        TestWMLClientWithKeras.logger.debug(details)

        self.assertTrue("published_models" in str(details))
        self.assertEqual(type(details), dict)

    def test_02_save_definition1(self):
        TestWMLClientWithKeras.logger.info("Save model definition ...")

        self.client.repository.DefinitionMetaNames.show()

        metadata = {
            self.client.repository.DefinitionMetaNames.NAME: "PyTorch model definition",
            self.client.repository.DefinitionMetaNames.AUTHOR_NAME: "John Smith",
            self.client.repository.DefinitionMetaNames.AUTHOR_EMAIL: "js@js.com",
            self.client.repository.DefinitionMetaNames.FRAMEWORK_NAME: "pytorch",
            self.client.repository.DefinitionMetaNames.FRAMEWORK_VERSION: "0.3",
            self.client.repository.DefinitionMetaNames.RUNTIME_NAME: "python",
            self.client.repository.DefinitionMetaNames.RUNTIME_VERSION: "3.5",
            self.client.repository.DefinitionMetaNames.EXECUTION_COMMAND: "python3 torch_check.py"
        }

        model_content_path = './artifacts/torch_train.zip'
        definition_details = self.client.repository.store_definition(training_definition=model_content_path, meta_props=metadata)
        TestWMLClientWithKeras.definition_1_url = self.client.repository.get_definition_url(definition_details)
        TestWMLClientWithKeras.definition_1_uid = self.client.repository.get_definition_uid(definition_details)
        TestWMLClientWithKeras.logger.info("Saved model definition url: " + str(TestWMLClientWithKeras.definition_1_url))

    def test_03_save_definition2(self):
        TestWMLClientWithKeras.logger.info("Save model definition ...")

        self.client.repository.DefinitionMetaNames.show()

        metadata = {
            self.client.repository.DefinitionMetaNames.NAME: "PyTorch model definition",
            self.client.repository.DefinitionMetaNames.AUTHOR_NAME: "John Smith",
            self.client.repository.DefinitionMetaNames.AUTHOR_EMAIL: "js@js.com",
            self.client.repository.DefinitionMetaNames.FRAMEWORK_NAME: "pytorch",
            self.client.repository.DefinitionMetaNames.FRAMEWORK_VERSION: "0.3",
            self.client.repository.DefinitionMetaNames.RUNTIME_NAME: "python",
            self.client.repository.DefinitionMetaNames.RUNTIME_VERSION: "3.5",
            self.client.repository.DefinitionMetaNames.EXECUTION_COMMAND: "python3 torch_check.py"
        }

        model_content_path = './artifacts/torch_train.zip'
        definition_details = self.client.repository.store_definition(training_definition=model_content_path, meta_props=metadata)
        TestWMLClientWithKeras.definition_2_url = self.client.repository.get_definition_url(definition_details)
        TestWMLClientWithKeras.definition_2_uid = self.client.repository.get_definition_uid(definition_details)
        TestWMLClientWithKeras.logger.info("Saved model definition url: " + str(TestWMLClientWithKeras.definition_2_url))

    def test_04_get_definition_details(self):
        TestWMLClientWithKeras.logger.info("Getting definition details ...")
        details = self.client.repository.get_definition_details()
        TestWMLClientWithKeras.logger.info(details)
        self.assertTrue('PyTorch model definition' in str(details))
        TestWMLClientWithKeras.logger.info("List definitions")
        self.client.repository.list_definitions()

    def test_05_save_experiment(self):
        metadata = {
                    self.client.repository.ExperimentMetaNames.NAME: "xxx",
                    self.client.repository.ExperimentMetaNames.AUTHOR_EMAIL: "js@js.com",
                    self.client.repository.ExperimentMetaNames.EVALUATION_METHOD: "binary",
                    self.client.repository.ExperimentMetaNames.EVALUATION_METRICS: [],
                    self.client.repository.ExperimentMetaNames.TRAINING_DATA_REFERENCE: {
                        "type": "s3",
                        "connection": {
                            "access_key_id": "xxx",
                            "secret_access_key": "xxx",
                            "endpoint_url": "https://xxx.ibm.com"
                        },
                        "source": {
                            "bucket": "xxx",
                            "key": "xxx"
                        }
                    },
                    self.client.repository.ExperimentMetaNames.TRAINING_RESULTS_REFERENCE: {
                        "type": "s3",
                        "connection": {
                            "access_key_id": "xxx",
                            "secret_access_key": "xxx",
                            "endpoint_url": "https://xxx.ibm.com"
                        },
                        "target": {
                            "bucket": "xxx",
                            "key": "xxx"
                        }
                    },
                    self.client.repository.ExperimentMetaNames.TRAINING_REFERENCES: [
                        {
                            "name": "xxx",
                            "training_definition_url": "https://xxx.ibm.com",
                            "command": "test",
                            "compute_configuration": {
                                "name": "k80"
                            }
                        }
                    ]
                }

        print(get_cos_training_data_reference(self.bucket_names))
        print(get_cos_training_results_reference(self.bucket_names))
        experiment_details = self.client.repository.store_experiment(meta_props=metadata)

        TestWMLClientWithKeras.experiment_uid = self.client.repository.get_experiment_uid(experiment_details)

        experiment_specific_details = self.client.repository.get_experiment_details(TestWMLClientWithKeras.experiment_uid)
        self.assertTrue(TestWMLClientWithKeras.experiment_uid in str(experiment_specific_details))

    def test_06_update_experiment(self):
        metadata = {
            self.client.repository.ExperimentMetaNames.NAME: "my_experiment",
            self.client.repository.ExperimentMetaNames.DESCRIPTION: "mnist best model",
            self.client.repository.ExperimentMetaNames.AUTHOR_NAME: "John Smith",
            self.client.repository.ExperimentMetaNames.EVALUATION_METHOD: "multiclass",
            self.client.repository.ExperimentMetaNames.EVALUATION_METRICS: ["accuracy"],
            self.client.repository.ExperimentMetaNames.TRAINING_DATA_REFERENCE: get_cos_training_data_reference(
                self.bucket_names),
            self.client.repository.ExperimentMetaNames.TRAINING_RESULTS_REFERENCE: get_cos_training_results_reference(
                self.bucket_names),
            self.client.repository.ExperimentMetaNames.TRAINING_REFERENCES: [
                {
                    "name": "mnist_nn",
                    "training_definition_url": TestWMLClientWithKeras.definition_1_url,
                    "compute_configuration": {"name": "k80"}
                },
                {
                    "name": "mnist_cnn",
                    "training_definition_url": TestWMLClientWithKeras.definition_2_url,
                    "compute_configuration": {"name": "k80"}
                },
            ]
        }

        experiment_details = self.client.repository.update_experiment(TestWMLClientWithKeras.experiment_uid, metadata)
        self.assertTrue('my_experiment' in str(experiment_details))
        print(experiment_details)
        self.assertTrue('xxx' not in str(experiment_details))

    def test_07_get_experiment_details(self):
        details = self.client.repository.get_experiment_details()
        self.assertTrue(TestWMLClientWithKeras.experiment_uid in str(details))

        details2 = self.client.repository.get_experiment_details(TestWMLClientWithKeras.experiment_uid)
        self.assertTrue(TestWMLClientWithKeras.experiment_uid in str(details2))

    def test_08_run_experiment(self):
        created_experiment_run_details = self.client.experiments.run(TestWMLClientWithKeras.experiment_uid, asynchronous=True)
        self.assertIsNotNone(created_experiment_run_details)
        TestWMLClientWithKeras.experiment_run_uid = Experiments.get_run_uid(created_experiment_run_details)

    def test_09_monitor(self):
        captured_output = io.StringIO()  # Create StringIO object
        sys.stdout = captured_output  # and redirect stdout.
        self.client.experiments.monitor_logs(TestWMLClientWithKeras.experiment_run_uid)  # Call function.
        sys.stdout = sys.__stdout__  # Reset redirect.
        self.assertTrue('training-' in captured_output.getvalue())

    def test_10_get_status(self):
        while True:
            time.sleep(3)
            status = self.client.experiments.get_status(TestWMLClientWithKeras.experiment_run_uid)
            if status['state'] == 'completed' or status['state'] == 'error' or status['state'] == 'canceled':
                break
        self.assertIsNotNone(status)
        self.assertTrue(status['state'] == 'completed')

    def test_11_get_all_experiments_runs_details(self):
        details = self.client.experiments.get_details()
        self.assertIsNotNone(details)

    def test_12_get_experiment_details(self):
        details = self.client.experiments.get_details(TestWMLClientWithKeras.experiment_uid)
        self.assertIsNotNone(details)

    def test_13_get_experiment_run_details(self):
        details = self.client.experiments.get_run_details(TestWMLClientWithKeras.experiment_run_uid)
        self.assertIsNotNone(details)

        self.assertIsNotNone(self.client.experiments.get_training_runs(details))
        self.assertIsNotNone(self.client.experiments.get_training_uids(details))

    def test_14_list_experiments(self):
        captured_output = io.StringIO()  # Create StringIO object
        sys.stdout = captured_output  # and redirect stdout.
        self.client.experiments.list_runs()  # Call function.
        sys.stdout = sys.__stdout__  # Reset redirect.
        self.assertTrue(TestWMLClientWithKeras.experiment_uid in captured_output.getvalue())
        self.client.experiments.list_runs()  # Just to see values.
        captured_output = io.StringIO()  # Create StringIO object
        sys.stdout = captured_output  # and redirect stdout.
        self.client.repository.list_experiments()  # Call function.
        sys.stdout = sys.__stdout__  # Reset redirect.
        self.assertTrue(TestWMLClientWithKeras.experiment_uid in captured_output.getvalue())
        self.client.repository.list_experiments()  # Just to see values.

    def test_15_list_experiment_runs_for_experiment(self):
        captured_output = io.StringIO()  # Create StringIO object
        sys.stdout = captured_output  # and redirect stdout.
        self.client.experiments.list_runs(TestWMLClientWithKeras.experiment_uid)  # Call function.
        sys.stdout = sys.__stdout__  # Reset redirect.
        self.assertTrue(TestWMLClientWithKeras.experiment_uid in captured_output.getvalue())
        self.client.experiments.list_runs(TestWMLClientWithKeras.experiment_uid)  # Just to see values.

    # def test_16_create_deployment(self):
    #     TestWMLClientWithCaffe.logger.info("Create deployment")
    #     TestWMLClientWithCaffe.train_run_uid = self.client.experiments.get_details(TestWMLClientWithCaffe.experiment_uid)['resources'][0]['entity']['training_statuses'][0]['training_guid']
    #     TestWMLClientWithCaffe.trained_model_uid = self.client.repository.store_model(TestWMLClientWithCaffe.train_run_uid, "test pytorch")
    #     deployment_details = self.client.deployments.create(artifact_uid=TestWMLClientWithCaffe.trained_model_uid, name="Test deployment", asynchronous=False)
    #     TestWMLClientWithCaffe.logger.debug("Deployment details: " + str(deployment_details))
    #     TestWMLClientWithCaffe.deployment_uid = self.client.deployments.get_uid(deployment_details)
    #     TestWMLClientWithCaffe.scoring_url = self.client.deployments.get_scoring_url(deployment_details)
    #     self.assertTrue('online' in str(TestWMLClientWithCaffe.scoring_url))
    #
    # def test_17_scoring(self):
    #     TestWMLClientWithCaffe.logger.info("Score model")
    #     scoring_data = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    #                              0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    #                              0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    #                              0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    #                              0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    #                              0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    #                              0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    #                              0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    #                              0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    #                              0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    #                              0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    #                              0, 0, 0, 0, 0, 0, 0, 0, 0, 3, 18, 18, 18,
    #                              126, 136, 175, 26, 166, 255, 247, 127, 0, 0, 0, 0, 0,
    #                              0, 0, 0, 0, 0, 0, 0, 30, 36, 94, 154, 170, 253,
    #                              253, 253, 253, 253, 225, 172, 253, 242, 195, 64, 0, 0, 0,
    #                              0, 0, 0, 0, 0, 0, 0, 0, 49, 238, 253, 253, 253,
    #                              253, 253, 253, 253, 253, 251, 93, 82, 82, 56, 39, 0, 0,
    #                              0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 18, 219, 253,
    #                              253, 253, 253, 253, 198, 182, 247, 241, 0, 0, 0, 0, 0,
    #                              0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    #                              80, 156, 107, 253, 253, 205, 11, 0, 43, 154, 0, 0, 0,
    #                              0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    #                              0, 0, 0, 14, 1, 154, 253, 90, 0, 0, 0, 0, 0,
    #                              0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    #                              0, 0, 0, 0, 0, 0, 0, 139, 253, 190, 2, 0, 0,
    #                              0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    #                              0, 0, 0, 0, 0, 0, 0, 0, 0, 11, 190, 253, 70,
    #                              0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    #                              0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 35,
    #                              241, 225, 160, 108, 1, 0, 0, 0, 0, 0, 0, 0, 0,
    #                              0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    #                              0, 0, 81, 240, 253, 253, 119, 25, 0, 0, 0, 0, 0,
    #                              0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    #                              0, 0, 0, 0, 0, 45, 186, 253, 253, 150, 27, 0, 0,
    #                              0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    #                              0, 0, 0, 0, 0, 0, 0, 0, 16, 93, 252, 253, 187,
    #                              0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    #                              0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 249,
    #                              253, 249, 64, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    #                              0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 46, 130,
    #                              183, 253, 253, 207, 2, 0, 0, 0, 0, 0, 0, 0, 0,
    #                              0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 39, 148,
    #                              229, 253, 253, 253, 250, 182, 0, 0, 0, 0, 0, 0, 0,
    #                              0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 24, 114,
    #                              221, 253, 253, 253, 253, 201, 78, 0, 0, 0, 0, 0, 0,
    #                              0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 23, 66,
    #                              213, 253, 253, 253, 253, 198, 81, 2, 0, 0, 0, 0, 0,
    #                              0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 18, 171,
    #                              219, 253, 253, 253, 253, 195, 80, 9, 0, 0, 0, 0, 0,
    #                              0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 55, 172,
    #                              226, 253, 253, 253, 253, 244, 133, 11, 0, 0, 0, 0, 0,
    #                              0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    #                              136, 253, 253, 253, 212, 135, 132, 16, 0, 0, 0, 0, 0,
    #                              0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    #                              0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    #                              0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    #                              0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    #                              0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    #                              0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    #                              0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
    #                              0, 0, 0, 0]
    #
    #     scoring_payload = {'values': [scoring_data, scoring_data]}
    #     scores = self.client.deployments.score(TestWMLClientWithCaffe.scoring_url, scoring_payload)
    #     self.assertIsNotNone(scores)
    #
    # def test_18_delete_deployment(self):
    #     TestWMLClientWithCaffe.logger.info("Delete deployment")
    #     self.client.deployments.delete(TestWMLClientWithCaffe.deployment_uid)
    #
    # def test_19_delete_model(self):
    #     TestWMLClientWithCaffe.logger.info("Delete model")
    #     self.client.repository.delete(TestWMLClientWithCaffe.trained_model_uid)

    def test_20_delete_experiment_run(self):
        self.client.experiments.delete(TestWMLClientWithKeras.experiment_run_uid)

    def test_21_list(self):
        captured_output = io.StringIO()  # Create StringIO object
        sys.stdout = captured_output  # and redirect stdout.
        self.client.repository.list_definitions()  # Call function.
        sys.stdout = sys.__stdout__  # Reset redirect.
        self.assertTrue(TestWMLClientWithKeras.definition_1_uid in captured_output.getvalue())
        self.assertTrue(TestWMLClientWithKeras.definition_2_uid in captured_output.getvalue())
        self.client.repository.list_definitions() # just to see

        captured_output = io.StringIO()  # Create StringIO object
        sys.stdout = captured_output  # and redirect stdout.
        self.client.repository.list_experiments()  # Call function.
        sys.stdout = sys.__stdout__  # Reset redirect.
        self.assertTrue(TestWMLClientWithKeras.experiment_uid in captured_output.getvalue())
        self.client.repository.list_experiments()  # just to see

        self.client.repository.list()

    def test_22_delete_experiment(self):
        self.client.repository.delete(TestWMLClientWithKeras.experiment_uid)

    def test_23_delete_definitions(self):
        self.client.repository.delete(TestWMLClientWithKeras.definition_1_uid)
        self.client.repository.delete(TestWMLClientWithKeras.definition_2_uid)

    def test_24_list(self):
        captured_output = io.StringIO()  # Create StringIO object
        sys.stdout = captured_output  # and redirect stdout.
        self.client.repository.list()  # Call function.
        sys.stdout = sys.__stdout__  # Reset redirect.
        self.assertTrue(TestWMLClientWithKeras.definition_1_uid not in captured_output.getvalue())
        self.assertTrue(TestWMLClientWithKeras.definition_2_uid not in captured_output.getvalue())
        self.assertTrue(TestWMLClientWithKeras.experiment_uid not in captured_output.getvalue())
        self.client.repository.list()  # Just to see values.


if __name__ == '__main__':
    unittest.main()



################################################################################
#
# Licensed Materials - Property of IBM
# (C) Copyright IBM Corp. 2017
# US Government Users Restricted Rights - Use, duplication disclosure restricted
# by GSA ADP Schedule Contract with IBM Corp.
#
################################################################################


from .spark_artifact_loader import SparkArtifactLoader


class SparkPipelineModelLoader(SparkArtifactLoader):
    """
        Returns pipeline model instance associated with this model artifact.

        :return: pipeline model
        :rtype: pyspark.ml.PipelineModel
        """
    def model_instance(self):
        return self.load()


################################################################################
#
# Licensed Materials - Property of IBM
# (C) Copyright IBM Corp. 2017
# US Government Users Restricted Rights - Use, duplication disclosure restricted
# by GSA ADP Schedule Contract with IBM Corp.
#
################################################################################


import logging

from ibm_watson_machine_learning.libs.repo.mlrepository.artifact_reader import ArtifactReader

logger = logging.getLogger('LibraryArtifactReader')


class LibrariesArtifactReader(ArtifactReader):
    def __init__(self, compressed_archive_path):
        self.library = compressed_archive_path

    def read(self):
        return self._open_stream()

    # This is a no. op. for LibraryZipFileReader as we do not want to delete the
    # archive file.
    def close(self):
        pass

    def _open_stream(self):
        return open(self.library, 'rb')




################################################################################
#
# Licensed Materials - Property of IBM
# (C) Copyright IBM Corp. 2017
# US Government Users Restricted Rights - Use, duplication disclosure restricted
# by GSA ADP Schedule Contract with IBM Corp.
#
################################################################################


class ArtifactReader(object):
    """
    Template reader class used to read artifacts.
    """

    def read(self):
        """
        Returns stream object with content of pipeline/pipeline model.

        :return: binary stream
        """
        pass

    def close(self):
        """
        Closes stream to content.
        """
        pass



################################################################################
#
# Licensed Materials - Property of IBM
# (C) Copyright IBM Corp. 2017
# US Government Users Restricted Rights - Use, duplication disclosure restricted
# by GSA ADP Schedule Contract with IBM Corp.
#
################################################################################


from ibm_watson_machine_learning.libs.repo.mlrepository import MetaNames, MetaProps, WmlExperimentArtifact
from ibm_watson_machine_learning.libs.repo.util.exceptions import MetaPropMissingError


class ExperimentArtifact(WmlExperimentArtifact):
    """
    Class of Experiment artifacts created with MLRepositoryCLient.

    """
    def __init__(self,  uid=None, name=None, meta_props=MetaProps({})):
        super(ExperimentArtifact, self).__init__(uid, name, meta_props)


        if meta_props.prop(MetaNames.EXPERIMENTS.PATCH_INPUT) is None:
            if meta_props.prop(MetaNames.EXPERIMENTS.SETTINGS) is None :
                 raise MetaPropMissingError('Value specified for "meta_props" does not contain value for '
                                         '"MetaNames.EXPERIMENTS.SETTINGS"')

            if meta_props.prop(MetaNames.EXPERIMENTS.TRAINING_REFERENCES) is None :
                raise MetaPropMissingError('Value specified for "meta_props" does not contain value for '
                                       '"MetaNames.EXPERIMENTS.TRAINING_REFERENCES"')

            if meta_props.prop(MetaNames.EXPERIMENTS.TRAINING_DATA_REFERENCE) is None :
                raise MetaPropMissingError('Value specified for "meta_props" does not contain value for '
                                       '"MetaNames.EXPERIMENTS.TRAINING_DATA_REFERENCE"')





import unittest
import time
from ibm_watson_machine_learning.utils.log_util import get_logger
from ibm_watson_machine_learning.experiments import Experiments
from preparation_and_cleaning import *


class TestWMLClientWithKeras(unittest.TestCase):
    deployment_uid = None
    definition_1_uid = None
    definition_2_uid = None
    definition_1_url = None
    definition_2_url = None
    experiment_uid = None
    experiment_run_uid = None
    experiment_run_details = None
    trained_model_uid = None
    train_run_id = None
    scoring_url = None
    logger = get_logger(__name__)

    @classmethod
    def setUpClass(self):
        TestWMLClientWithKeras.logger.info("Service Instance: setting up credentials")
        self.wml_credentials = get_wml_credentials()
        self.client = get_client()
        self.cos_resource = get_cos_resource()
        self.bucket_names = prepare_cos(self.cos_resource, data_code=BOSTON)

    @classmethod
    def tearDownClass(self):
        clean_cos(self.cos_resource, self.bucket_names)

    def test_01_service_instance_details(self):
        TestWMLClientWithKeras.logger.info("Check client ...")
        self.assertTrue(self.client.__class__.__name__ == 'APIClient')

        TestWMLClientWithKeras.logger.info("Getting instance details ...")
        details = self.client.service_instance.get_details()
        TestWMLClientWithKeras.logger.debug(details)

        self.assertTrue("published_models" in str(details))
        self.assertEqual(type(details), dict)

    def test_02_save_definition1(self):
        TestWMLClientWithKeras.logger.info("Save model definition ...")

        self.client.repository.DefinitionMetaNames.show()

        metadata = {
            self.client.repository.DefinitionMetaNames.NAME: "my_training_definition",
            self.client.repository.DefinitionMetaNames.DESCRIPTION: "my_description",
            self.client.repository.DefinitionMetaNames.AUTHOR_NAME: "John Smith",
            self.client.repository.DefinitionMetaNames.AUTHOR_EMAIL: "js@js.com",
            self.client.repository.DefinitionMetaNames.FRAMEWORK_NAME: "tensorflow",
            self.client.repository.DefinitionMetaNames.FRAMEWORK_VERSION: "1.5",
            self.client.repository.DefinitionMetaNames.RUNTIME_NAME: "python",
            self.client.repository.DefinitionMetaNames.RUNTIME_VERSION: "3.5",
            self.client.repository.DefinitionMetaNames.EXECUTION_COMMAND: "python3 keras_boston.py"
        }

        model_content_path = './artifacts/keras_boston.zip'
        definition_details = self.client.repository.store_definition(training_definition=model_content_path, meta_props=metadata)
        TestWMLClientWithKeras.definition_1_url = self.client.repository.get_definition_url(definition_details)
        TestWMLClientWithKeras.definition_1_uid = self.client.repository.get_definition_uid(definition_details)
        TestWMLClientWithKeras.logger.info("Saved model definition url: " + str(TestWMLClientWithKeras.definition_1_url))

    def test_03_get_definition_details(self):
        TestWMLClientWithKeras.logger.info("Getting definition details ...")
        details = self.client.repository.get_definition_details()
        TestWMLClientWithKeras.logger.info(details)
        self.assertTrue('my_training_definition' in str(details))
        TestWMLClientWithKeras.logger.info("List definitions")
        self.client.repository.list_definitions()

    def test_04_save_experiment(self):
        metadata = {
            self.client.repository.ExperimentMetaNames.NAME: "my_experiment",
            self.client.repository.ExperimentMetaNames.DESCRIPTION: "mnist best model",
            self.client.repository.ExperimentMetaNames.AUTHOR_NAME: "John Smith",
            self.client.repository.ExperimentMetaNames.EVALUATION_METHOD: "multiclass",
            self.client.repository.ExperimentMetaNames.EVALUATION_METRICS: ["accuracy"],
            self.client.repository.ExperimentMetaNames.TRAINING_DATA_REFERENCE: get_cos_training_data_reference(
                self.bucket_names),
            self.client.repository.ExperimentMetaNames.TRAINING_RESULTS_REFERENCE: get_cos_training_results_reference(
                self.bucket_names),
            self.client.repository.ExperimentMetaNames.TRAINING_REFERENCES: [
                {
                    "name": "mnist_nn",
                    "training_definition_url": TestWMLClientWithKeras.definition_1_url,
                    "compute_configuration": {"name": "k80"}
                }
            ]
        }

        print(get_cos_training_data_reference(self.bucket_names))
        print(get_cos_training_results_reference(self.bucket_names))
        experiment_details = self.client.repository.store_experiment(meta_props=metadata)

        TestWMLClientWithKeras.experiment_uid = self.client.repository.get_experiment_uid(experiment_details)

        experiment_specific_details = self.client.repository.get_experiment_details(TestWMLClientWithKeras.experiment_uid)
        self.assertTrue(TestWMLClientWithKeras.experiment_uid in str(experiment_specific_details))

    def test_05_get_experiment_details(self):
        details = self.client.repository.get_experiment_details()
        self.assertTrue(TestWMLClientWithKeras.experiment_uid in str(details))

        details2 = self.client.repository.get_experiment_details(TestWMLClientWithKeras.experiment_uid)
        self.assertTrue(TestWMLClientWithKeras.experiment_uid in str(details2))

    def test_06_run_experiment(self):
        created_experiment_run_details = self.client.experiments.run(TestWMLClientWithKeras.experiment_uid, asynchronous=False)
        self.assertIsNotNone(created_experiment_run_details)
        TestWMLClientWithKeras.experiment_run_uid = Experiments.get_run_uid(created_experiment_run_details)
        print("Experiment run id:\n{}".format(TestWMLClientWithKeras.experiment_run_uid))

    def test_07_get_status(self):
        start_time = time.time()
        diff_time = start_time - start_time
        while True and diff_time < 60 * 10:
            time.sleep(3)
            status = self.client.experiments.get_status(TestWMLClientWithKeras.experiment_run_uid)
            if status['state'] == 'completed' or status['state'] == 'error' or status['state'] == 'canceled':
                break
            diff_time = time.time() - start_time
        self.assertIsNotNone(status)
        self.assertTrue(status['state'] == 'completed')

    def test_08_get_experiment_details(self):
        details = self.client.experiments.get_details(TestWMLClientWithKeras.experiment_uid)
        print("Experiment details:\n{}".format(details))
        self.assertIsNotNone(details)

    def test_09_get_experiment_run_details(self):
        TestWMLClientWithKeras.experiment_run_details = self.client.experiments.get_run_details(TestWMLClientWithKeras.experiment_run_uid)
        print("Experiment run details:\n{}".format(TestWMLClientWithKeras.experiment_run_details))
        self.assertIsNotNone(TestWMLClientWithKeras.experiment_run_details)
        self.assertIsNotNone(self.client.experiments.get_training_runs(TestWMLClientWithKeras.experiment_run_details))

    def test_10_store_experiment_as_model(self):
        training_run_uids = self.client.experiments.get_training_uids(TestWMLClientWithKeras.experiment_run_details)
        training_run_uid = training_run_uids[0]

        model_props = {
            self.client.repository.ModelMetaNames.NAME: "Keras Experiment Model",
        }
        model_details = self.client.repository.store_model(training_run_uid, meta_props=model_props)
        TestWMLClientWithKeras.trained_model_uid = self.client.repository.get_model_uid(model_details)

    def test_11_create_learning_system(self):
        self.client.learning_system.ConfigurationMetaNames.show()

        meta_prop = {
            self.client.learning_system.ConfigurationMetaNames.FEEDBACK_DATA_REFERENCE: get_cos_training_data_reference(self.bucket_names),
            self.client.learning_system.ConfigurationMetaNames.TRAINING_RESULTS_REFERENCE: get_cos_training_results_reference(self.bucket_names),
            self.client.learning_system.ConfigurationMetaNames.MIN_FEEDBACK_DATA_SIZE: 10,
            self.client.learning_system.ConfigurationMetaNames.AUTO_RETRAIN: "always",
            self.client.learning_system.ConfigurationMetaNames.AUTO_REDEPLOY: "always",
            self.client.learning_system.ConfigurationMetaNames.COMPUTE_CONFIGURATION: {
              "name": "k80x2",
              "nodes": 1
            }
        }

        details = self.client.learning_system.setup(TestWMLClientWithKeras.trained_model_uid, meta_prop)
        self.assertIsNotNone(details)

    def test_12_create_deployment(self):
        deployment_details = self.client.deployments.create(artifact_uid=TestWMLClientWithKeras.trained_model_uid, name="Test deployment", asynchronous=False)
        TestWMLClientWithKeras.logger.debug("Deployment details: " + str(deployment_details))
        TestWMLClientWithKeras.deployment_uid = self.client.deployments.get_uid(deployment_details)
        TestWMLClientWithKeras.scoring_url = self.client.deployments.get_scoring_url(deployment_details)
        self.assertTrue('online' in str(TestWMLClientWithKeras.scoring_url))

    def test_13_list(self):
        self.client.learning_system.list()  # Call function.

    def test_14_run_learning_system(self):
        run_details = self.client.learning_system.run(TestWMLClientWithKeras.trained_model_uid, asynchronous=False)
        TestWMLClientWithKeras.run_uid = self.client.learning_system.get_run_uid(run_details)
        url = self.client.learning_system.get_run_href(run_details)
        self.assertIsNotNone(url)

    def test_15_get_runs(self):
        runs = self.client.learning_system.get_runs(TestWMLClientWithKeras.trained_model_uid)
        self.assertIsNotNone(runs)
        self.assertTrue(TestWMLClientWithKeras.run_uid in str(runs))

    def test_16_get_deployment_status(self):
        print(self.client.deployments.get_status(TestWMLClientWithKeras.deployment_uid))

    def test_17_get_metrics(self):
        metrics = self.client.learning_system.get_metrics(TestWMLClientWithKeras.trained_model_uid)
        print(metrics)
        self.assertTrue('values' in str(metrics))

    def test_18_list_metrics(self):
        self.client.learning_system.list_metrics(TestWMLClientWithKeras.trained_model_uid)

    def test_19_list_runs(self):
        self.client.learning_system.list_runs(TestWMLClientWithKeras.trained_model_uid)

    def test_20_list_all_runs(self):
        self.client.learning_system.list_runs()

    def test_21_get_run_details(self):
        details = self.client.learning_system.get_run_details(TestWMLClientWithKeras.run_uid)
        self.assertTrue('COMPLETED' in str(details))

    def test_22_delete_deployment(self):
        TestWMLClientWithKeras.logger.info("Delete deployment")
        self.client.deployments.delete(TestWMLClientWithKeras.deployment_uid)

    def test_23_delete_model(self):
        TestWMLClientWithKeras.logger.info("Delete model")
        self.client.repository.delete(TestWMLClientWithKeras.trained_model_uid)

    def test_24_delete_experiment_run(self):
        self.client.experiments.delete(TestWMLClientWithKeras.experiment_run_uid)

    def test_25_delete_experiment(self):
        self.client.repository.delete(TestWMLClientWithKeras.experiment_uid)

    def test_26_delete_definition(self):
        self.client.repository.delete(TestWMLClientWithKeras.definition_1_uid)


if __name__ == '__main__':
    unittest.main()



__all__ = [
    'get_project',
    'get_wmls_credentials_and_space_ids',
    'get_iam_token',
    'get_project_environments_details',
    'create_spark_header',
    'change_client_get_headers'
]

import base64
import json
from typing import Any, Dict, Tuple, List, TYPE_CHECKING

import requests

from ibm_watson_machine_learning.href_definitions import IAM_TOKEN_API, HrefDefinitions
from ibm_watson_machine_learning.wml_client_error import ApiRequestFailure
from .errors import NotInWatsonStudio, CredentialsNotFound

if TYPE_CHECKING:
    from ibm_watson_machine_learning.workspace import WorkSpace


def get_project() -> Any:
    """Try to import project_lib and get user corresponding project."""
    try:
        from project_lib import Project

    except ModuleNotFoundError:
        raise NotInWatsonStudio(reason="You are not in Watson Studio or Watson Studio Desktop environment. "
                                       "Cannot access to project metadata.")

    try:
        access = Project.access()

    except RuntimeError:
        raise CredentialsNotFound(reason="Your WSD environment does not have correctly configured "
                                         "connection to WML Server or you are not in WSD environment. "
                                         "In that case, please provide WMLS credentials and space_id.")

    return access


def get_wmls_credentials_and_space_ids(local_path: str) -> Tuple[List[Dict], List[str]]:
    """
    Parse project.json file and get WMLS credentials associated with WSD project.

    Parameters
    ----------
    local_path: str, required
        Path to project.json file with project configuration in WSD.

    Returns
    -------
    Two lists with WMLS credentials and corresponding space_ids.
    """
    try:
        with open(local_path, 'r') as f:
            data = json.load(f)

        credentials = [instance['credentials'] for instance in data['compute']]
        for instance in credentials:
            instance['version'] = "2.0"
            instance['instance_id'] = "wml_local"
            instance['password'] = base64.decodebytes(bytes(instance['password'].encode())).decode()

        space_ids = [instance['properties']['space_guid'] for instance in data['compute']]

    except (FileNotFoundError, KeyError):
        raise CredentialsNotFound(reason="Your WSD environment does not have correctly configured "
                                         "connection to WML Server or you are not in WSD environment. "
                                         "In that case, please provide WMLS credentials and space_id.")

    return credentials, space_ids


def get_iam_token(workspace: 'WorkSpace') -> str:
    """
    Fetch IAM token for given workspace.

    Parameters
    ----------
    workspace: WorkSpace, required

    Returns
    -------
    String with IAM token
    """
    _href_definitions = HrefDefinitions(workspace.wml_client.wml_credentials)
    try:
        iam_api_key = workspace.wml_credentials['iam_api_key']

    except KeyError:
        raise AttributeError("You need to pass an IAM api key in WML credentials "
                             "to be able to run multiple files preprocessing step.")

    headers = {
        'Content-Type': 'application/x-www-form-urlencoded',
        'Authorization': 'Basic Yng6Yng='
    }
    data = f"apikey={IAM_TOKEN_API.format(iam_api_key)}"

    response = requests.post(
        url=_href_definitions.get_iam_token_url(),
        data=data,
        headers=headers
    )

    if response.status_code == 200:
        token = response.json().get(u'access_token')
    else:
        raise ApiRequestFailure(u'Error during getting IAM Token.', response)
    return token


def create_ws_env_url(workspace: 'WorkSpace') -> str:
    """Make url to Watson Studio Environments.

    Parameters
    ----------
    workspace: WorkSpace, required

    Returns
    -------
    str with url
    """
    if workspace.project_id is None:
        raise ValueError("Missing project_id. It is required to pass project_id from Watson Studio "
                         "for Data Join + AutoAI training and deployment.")

    if 'test' in workspace.wml_credentials['url']:
        url = f'https://api.dataplatform.dev.cloud.ibm.com/v2/environments?project_id={workspace.project_id}'

    else:
        url = f'https://api.dataplatform.cloud.ibm.com/v2/environments?project_id={workspace.project_id}'

    return url


def get_project_environments_details(workspace: 'WorkSpace') -> Dict:
    """
    Fetch environments details related to user project in Watson Studio.

    Parameters
    ----------
    workspace: WorkSpace, required

    Returns
    -------
    Dictionary with environments details.
    """
    token = get_iam_token(workspace)
    url = create_ws_env_url(workspace)

    headers = {
        'Authorization': f"Bearer {token}",
        'X-WML-User-Client': 'PythonClient',
        'Content-Type': 'application/json'
    }

    response = requests.get(
        url=url,
        headers=headers
    )

    if response.status_code == 200:
        return response.json()

    else:
        raise ApiRequestFailure(u'Getting project environment details failed.', response)


def create_spark_header(environments_details: Dict, workspace: 'WorkSpace') -> str:
    """Creates header with spark environment."""
    env_guid = None

    for resource in environments_details['resources']:
        if 'spark24scala' == resource['metadata']['name']:
            env_guid = resource['metadata']['asset_id']

    if env_guid is None:
        raise ValueError("Cannot find spark environment guid.")

    token = get_iam_token(workspace)
    url = create_ws_env_url(workspace)
    url = f"{url.split('?project_id')[0]}/{env_guid}/access_info?project_id={workspace.project_id}"

    headers = {
        'Authorization': f"Bearer {token}",
        'X-WML-User-Client': 'PythonClient',
        'Content-Type': 'application/json'
    }

    response = requests.get(
        url=url,
        headers=headers
    )

    if response.status_code == 200:
        data = response.json()

    else:
        raise ApiRequestFailure(u'Getting spark access details failed.', response)

    spark_header = {
        "credentials": {
            "host": data['entity']['url'].split('/')[2],
            "instance_id": data['entity']['instance_id'],
            "api_key": data['entity']['api_key']
        },
        "kernel_id": data['entity']['url'].split('/')[4]
    }

    # note: encode spark header with base64
    spark_header = json.dumps(spark_header)
    spark_header = base64.encodebytes(bytes(spark_header.encode()))
    spark_header = spark_header.replace(b'\n', b'')
    spark_header = spark_header.decode()
    # --- end note

    return spark_header


def change_client_get_headers(client: 'APIClient', _type: str, workspace: 'WorkSpace'):
    """Replace client _get_headers() method body with new one (including spark header)."""

    from types import MethodType
    from ibm_watson_machine_learning import APIClient

    def _get_headers_new(self, content_type='application/json', no_content_type=False,
                         spark_header=create_spark_header(get_project_environments_details(workspace), workspace)):
        if self.WSD:
            headers = {'X-WML-User-Client': 'PythonClient'}
            if self.project_id is not None:
                headers.update({'X-Watson-Project-ID': self.project_id})
            if not no_content_type:
                headers.update({'Content-Type': content_type})
        else:
            if self.proceed is True:
                token = "Bearer " + self.wml_credentials["token"]
            else:
                token = "Bearer " + self.service_instance._get_token()
            headers = {
                'Authorization': token,
                'X-WML-User-Client': 'PythonClient'
            }
            if self._is_IAM() or (self.service_instance._is_iam() is None):
                headers['ML-Instance-ID'] = self.wml_credentials['instance_id']

            headers.update({'x-wml-internal-switch-to-new-v4': "true"})
            if not self.ICP:
                # headers.update({'x-wml-internal-switch-to-new-v4': "true"})
                if self.project_id is not None:
                    headers.update({'X-Watson-Project-ID': self.project_id})

            if not no_content_type:
                headers.update({'Content-Type': content_type})

        headers["X-SparkHB-Service-Instance"] = spark_header

        return headers

    if _type == 'new':
        client._get_headers = MethodType(_get_headers_new, client)

    else:
        client._get_headers = MethodType(APIClient._get_headers, client)




class MetaPropMissingError(Exception):
    pass


class UnsupportedTFSerializationFormat(Exception):
    pass


class UnmatchedKerasVersion(Exception):
    pass

class InvalidCaffeModelArchive(Exception):
    pass


# coding: utf-8

"""
Copyright 2016 SmartBear Software

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

    Ref: https://github.com/swagger-api/swagger-codegen
"""

from pprint import pformat
from six import iteritems


class PatchOperationRuntimeSpec(object):
    """
    NOTE: This class is auto generated by the swagger code generator program.
    Do not edit the class manually.
    """
    def __init__(self, op=None, path=None, value=None, _from=None):
        """
        PatchOperationRuntimeSpec - a model defined in Swagger

        :param dict swaggerTypes: The key is attribute name
                                  and the value is attribute type.
        :param dict attributeMap: The key is attribute name
                                  and the value is json key in definition.
        """
        self.swagger_types = {
            'op': 'str',
            'path': 'str',
            'value': 'object',
            '_from': 'str'
        }

        self.attribute_map = {
            'op': 'op',
            'path': 'path',
            'value': 'value',
            '_from': 'from'
        }

        self._op = op
        self._path = path
        self._value = value
        self.__from = _from

    @property
    def op(self):
        """
        Gets the op of this PatchOperationRuntimeSpec.
        operation

        :return: The op of this PatchOperationRuntimeSpec.
        :rtype: str
        """
        return self._op

    @op.setter
    def op(self, op):
        """
        Sets the op of this PatchOperationRuntimeSpec.
        operation

        :param op: The op of this PatchOperationRuntimeSpec.
        :type: str
        """
        allowed_values = ["add", "remove", "replace", "move", "copy", "test"]
        if op not in allowed_values:
            raise ValueError(
                "Invalid value for `op`, must be one of {0}"
                .format(allowed_values)
            )
        self._op = op

    @property
    def path(self):
        """
        Gets the path of this PatchOperationRuntimeSpec.
        path

        :return: The path of this PatchOperationRuntimeSpec.
        :rtype: str
        """
        return self._path

    @path.setter
    def path(self, path):
        """
        Sets the path of this PatchOperationRuntimeSpec.
        path

        :param path: The path of this PatchOperationRuntimeSpec.
        :type: str
        """
        self._path = path

    @property
    def value(self):
        """
        Gets the value of this PatchOperationRuntimeSpec.


        :return: The value of this PatchOperationRuntimeSpec.
        :rtype: object
        """
        return self._value

    @value.setter
    def value(self, value):
        """
        Sets the value of this PatchOperationRuntimeSpec.


        :param value: The value of this PatchOperationRuntimeSpec.
        :type: object
        """
        self._value = value

    @property
    def _from(self):
        """
        Gets the _from of this PatchOperationRuntimeSpec.


        :return: The _from of this PatchOperationRuntimeSpec.
        :rtype: str
        """
        return self.__from

    @_from.setter
    def _from(self, _from):
        """
        Sets the _from of this PatchOperationRuntimeSpec.


        :param _from: The _from of this PatchOperationRuntimeSpec.
        :type: str
        """
        self.__from = _from

    def to_dict(self):
        """
        Returns the model properties as a dict
        """
        result = {}

        for attr, _ in iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """
        Returns the string representation of the model
        """
        return pformat(self.to_dict())

    def __repr__(self):
        """
        For `print` and `pprint`
        """
        return self.to_str()

    def __eq__(self, other):
        """
        Returns true if both objects are equal
        """
        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """
        Returns true if both objects are not equal
        """
        return not self == other




from ibm_watson_machine_learning.experiment.autoai.runs.auto_pipelines_runs import AutoPipelinesRuns
from ibm_watson_machine_learning.experiment.autoai.runs.local_auto_pipelines_runs import LocalAutoPipelinesRuns



from abc import ABC, abstractmethod

__all__ = [
    "BaseExperiment"
]


class BaseExperiment(ABC):
    """
    Base abstract class for Experiment.
    """

    @abstractmethod
    def runs(self, *, filter: str):
        """Get the historical runs but with WML Pipeline name filter.

        Parameters
        ----------
        filter: str, required
            WML Pipeline name to filter the historical runs.
        """
        pass



################################################################################
#
# Licensed Materials - Property of IBM
# (C) Copyright IBM Corp. 2017
# US Government Users Restricted Rights - Use, duplication disclosure restricted
# by GSA ADP Schedule Contract with IBM Corp.
#
################################################################################

from configparser import ConfigParser
import logging

class LogConfig:
    specific_log_level = logging.WARN
    filename = None

def get_logger(name, specific_log_level=None):
    default_log_level = LogConfig.specific_log_level

    config = ConfigParser()
    config.read("./config.ini")

    try:
        filename = config.get("DEFAULT", 'log-filename')
    except:
        filename = LogConfig.filename

    log_level = default_log_level

    try:
        env_log_level = getattr(logging, config.get("DEFAULT", 'log'), None)
    except:
        env_log_level = None

    if env_log_level is not None:
        log_level = env_log_level

    if specific_log_level is not None:
        log_level = specific_log_level

    if not isinstance(log_level, int):
        raise ValueError('Invalid log level: %s' % log_level)

    logger = logging.getLogger(name)
    logging.basicConfig(level=log_level, format="%(asctime)s - %(name)s - %(levelname)s - %(message)s", filename=filename)
    logger.setLevel(log_level)
    return logger



__all__ = [
    'DataJoinPipeline',
    'OBMPipelineGraphBuilder'
]

import re
from operator import itemgetter
from collections import defaultdict
from typing import Tuple, List, Union, Dict, TYPE_CHECKING
from graphviz import Digraph
from .templates import pretty_print_template
from ..utils.autoai.utils import is_ipython, check_graphviz_binaries
from ..utils.autoai.enums import VisualizationTypes

if TYPE_CHECKING:
    from ..helpers.connections.connections import DataConnection


class DataJoinPipeline:
    """
    Class representing abstract data join pipeline.

    Parameters
    ----------
    preprocessed_data_connection: DataConnection, required
        Populated DataConnection with preprocessed data join information.
    optimizer: RemoteAutoPipeline, required
        Optimizer with initialized pipeline metadata, required for running an OBM training in predict method.
    """
    VisualizationTypes = VisualizationTypes

    def __init__(self, preprocessed_data_connection: 'DataConnection', optimizer: 'RemoteAutoPipelines') -> None:
        self._pipeline_json = preprocessed_data_connection._download_obm_json_from_cos()['Pipeline']
        self._obm_pipeline_graph_builder = OBMPipelineGraphBuilder(self._pipeline_json)
        self._graph_json, self._graph = self._obm_pipeline_graph_builder.build_graph()
        self._optimizer = optimizer

    @check_graphviz_binaries
    def visualize(self, astype: 'VisualizationTypes' = VisualizationTypes.INPLACE) -> None:
        """Display graph in the notebook or as a rendered image.

        Parameters
        ----------
        astype: VisualizationTypes, optional
            Specify a type of visualization for this graph. Default is VisualizationTypes.INPLACE
            (when in notebook env, picture will be displayed in the cell output).
            VisualizationTypes.PDF indicates to render a pdf document with visualization.
        """
        if is_ipython():
            if astype == VisualizationTypes.PDF:
                self._graph.render(view=True)

            else:
                import IPython.display
                IPython.display.display(self._graph)

        else:
            self._graph.render(view=True)

    # def predict(self,
    #             *,
    #             training_data_reference: List['DataConnection'],
    #             training_results_reference: 'DataConnection' = None,
    #             background_mode=False) -> 'DataFrame':
    #     """
    #     Run an OBM training process on top of the training data referenced by DataConnection.
    #
    #     Parameters
    #     ----------
    #     training_data_reference: List[DataConnection], required
    #         Data storage connection details to inform where training data is stored.
    #
    #     training_results_reference: DataConnection, optional
    #         Data storage connection details to store pipeline training results. Not applicable on CP4D.
    #
    #     background_mode: bool, optional
    #         Indicator if fit() method will run in background (async) or (sync).
    #
    #     Returns
    #     -------
    #     pandas.DataFrame contains dataset from remote data storage.
    #     """
    #     self._optimizer.fit(training_data_reference=training_data_reference,
    #                         training_results_reference=training_results_reference,
    #                         background_mode=background_mode)
    #
    #     return self._optimizer.get_preprocessed_data_connection().read()

    def pretty_print(self, ipython_display: bool = False) -> None:
        """
        Prints code which generates OBM data preprocessing.

        Parameters
        ----------
        ipython_display: bool, optional
            If method executed in jupyter notebooks/ipython set this flag to true in order to get syntax highlighting.
        """
        params = self._optimizer.get_params()
        data_join_graph = params["data_join_graph"]

        nodes = ""
        for node in data_join_graph.nodes:
            if node.table.timestamp_format and node.table.timestamp_column_name:
                nodes += f"data_join_graph.node(name=\"{node.table.name}\"," \
                         f" timestamp_column_name=\"{node.table.timestamp_column_name}\"," \
                         f" timestamp_format=\"{node.timestamp_format}\")\n"
            else:
                nodes += f"data_join_graph.node(name=\"{node.table.name}\")\n"

        edges = "\n".join([f"data_join_graph.edge(from_node=\"{edge.from_node}\", to_node=\"{edge.to_node}\",\n"
                           f"\t\t\t\t\t from_column={edge.from_column}, to_column={edge.to_column})"
                           for edge in data_join_graph.edges])


        join_indices = [it - 1 for it in self._obm_pipeline_graph_builder.get_join_iterations()]
        paths = "\n".join(["# - " + re.search(r"\[(.*)\]", msg["message"]["text"]).group()
                           for msg in itemgetter(*join_indices)(self._pipeline_json)])
        nb_of_features = re.search(r"\d", self._pipeline_json[-1]["message"]["text"]).group()

        pretty_print = pretty_print_template.template.format(
                        nodes=nodes, edges=edges,
                        name=f"\"{params['name']}\"",
                        prediction_type=f"\"{params['prediction_type']}\"",
                        prediction_column=f"\"{params['prediction_column']}\"",
                        scoring=f"\"{params['scoring']}\"",
                        paths=paths,
                        nb_of_features=nb_of_features)

        if ipython_display:
            import IPython.display
            markdown = IPython.display.Markdown(f'```python\n{pretty_print}\n```')
            IPython.display.display(markdown)
        else:
            print(pretty_print)


class OBMPipelineGraphBuilder:
    """
    Class for extracting particular elements from OBM output json with pipeline description.

    Parameters
    ----------
    pipeline_json: dict, required
        Dictionary with loaded obm.json file.
    """
    def __init__(self, pipeline_json: dict) -> None:
        self.pipeline_json = pipeline_json
        self.last_non_join_iteration = self.get_last_non_join_iteration()
        self.selection_iteration = self.get_selection_iteration()
        self.join_iterations = self.get_join_iterations()
        self.graph_json = {'nodes': [], 'edges': defaultdict(set)}
        self.graph = Digraph(comment='Data Preprocessing Steps Graph',
                             node_attr={'color': 'lightblue2', 'style': 'filled'})

    @staticmethod
    def get_step_details(msg_json) -> Tuple[str, int, str]:
        """
        Getting particular step name, iteration number and step description.
        """
        name = msg_json['feature_engineering_components']['obm'][0]['step_name'].split(':')[1]
        iteration = msg_json['feature_engineering_components']['obm'][0]['iteration']
        text = msg_json['message']['text']
        return name, iteration, text

    def get_step_types(self) -> List[str]:
        """For all steps return their types."""
        return [message['feature_engineering_components']['obm'][0]['step_type'] for message in self.pipeline_json]

    def get_last_non_join_iteration(self) -> int:
        """Returns a number of the last step before join."""
        return self.get_step_types().index('join')

    def get_selection_iteration(self) -> int:
        """Returns feature selection step number."""
        return self.get_step_types().index('feature selection') + 1

    def get_join_iterations(self) -> List[int]:
        """Returns list of join step numbers."""
        steps_types = [message['feature_engineering_components']['obm'][0]['step_type'] for message in self.pipeline_json]
        return [i + 1 for i, x in enumerate(steps_types) if x == "join"]

    @staticmethod
    def get_join_extractors(msg_json: dict) -> Union[dict, None]:
        """Returns extractors if exist."""
        return msg_json['feature_engineering_components']['obm'][0].get('extractors')

    @staticmethod
    def get_extractor_columns(extractor_json: dict) -> List[str]:
        """Returns all columns names from particular extractor."""
        return extractor_json['columns']

    @staticmethod
    def get_extractor_transformations(extractor_json: dict) -> Dict[str, str]:
        """Returns a dictionary with all transformations names."""
        return extractor_json['transformations']

    def build_extractors_subgraph(self, msg_json: dict, join_iteration: int) -> None:
        """Creates sub-graph for extractors representation."""
        extractors = self.get_join_extractors(msg_json)
        join_iteration = str(join_iteration)

        if extractors is not None:
            for ext, i in zip(extractors, range(len(extractors))):
                self.graph.attr('node', color='lightgray')
                ext_index = join_iteration + str(i)

                self.graph.node(ext_index, ext)
                self.graph.edge(join_iteration, ext_index)
                self.graph_json['nodes'].append(ext_index)
                self.graph_json['edges'][join_iteration].add(ext_index)

                columns = self.get_extractor_columns(extractors[ext])
                transformations = self.get_extractor_transformations(extractors[ext])

                for j, column in enumerate(columns):
                    self.graph.attr('node', color='lightgreen')
                    col_index = join_iteration + str(i) + str(j)

                    self.graph.node(col_index, column)
                    self.graph.edge(ext_index, col_index)
                    self.graph_json['nodes'].append(col_index)
                    self.graph_json['edges'][ext_index].add(col_index)

                    self.graph.attr('node', color='lightblue2')
                    for transformation in transformations:
                        self.graph.edge(col_index, transformation)
                        self.graph.edge(transformation, str(self.selection_iteration))
                        self.graph_json['edges'][col_index].add(transformation)
                        self.graph_json['edges'][transformation].add(str(self.selection_iteration))

    def build_graph(self) -> Tuple[dict, 'Digraph']:
        """Creates a graphviz Digraph with pipeline representation."""
        for msg in self.pipeline_json:
            name, iteration, text = self.get_step_details(msg)
            self.graph.node(str(iteration), name, tooltip=text)
            self.graph_json['nodes'].append(str(iteration))

            if 1 < iteration <= self.last_non_join_iteration:
                self.graph.edge(str(iteration - 1), str(iteration))
                self.graph_json['edges'][str(iteration - 1)].add(str(iteration))
            elif iteration in self.join_iterations:
                self.graph.edge(str(self.last_non_join_iteration), str(iteration))
                self.graph_json['edges'][str(self.last_non_join_iteration)].add(str(iteration))
                self.build_extractors_subgraph(msg, iteration)
            elif iteration > self.selection_iteration:
                self.graph.edge(str(iteration - 1), str(iteration))
                self.graph_json['edges'][str(iteration - 1)].add(str(iteration))

        return self.graph_json, self.graph



from abc import ABC

__all__ = [
    "BaseLocation"
]


class BaseLocation(ABC):
    """
    Base class for storage Location.
    """

    def to_dict(self) -> dict:
        """Return a json dictionary representing this model."""
        return vars(self)

    def _get_file_size(self, **kwargs) -> 'int':
        raise NotImplementedError



# coding: utf-8

"""

    No descripton provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)

    
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from pprint import pformat
from six import iteritems
import re


class ErrorExperiments(object):
    """
    NOTE: This class is auto generated by the swagger code generator program.
    Do not edit the class manually.
    """
    def __init__(self, code=None, message=None, more_info=None, target=None):
        """
        ErrorExperiments - a model defined in Swagger

        :param dict swaggerTypes: The key is attribute name
                                  and the value is attribute type.
        :param dict attributeMap: The key is attribute name
                                  and the value is json key in definition.
        """
        self.swagger_types = {
            'code': 'str',
            'message': 'str',
            'more_info': 'str',
            'target': 'ErrorExperimentsTarget'
        }

        self.attribute_map = {
            'code': 'code',
            'message': 'message',
            'more_info': 'more_info',
            'target': 'target'
        }

        self._code = code
        self._message = message
        self._more_info = more_info
        self._target = target

    @property
    def code(self):
        """
        Gets the code of this ErrorExperiments.


        :return: The code of this ErrorExperiments.
        :rtype: str
        """
        return self._code

    @code.setter
    def code(self, code):
        """
        Sets the code of this ErrorExperiments.


        :param code: The code of this ErrorExperiments.
        :type: str
        """

        self._code = code

    @property
    def message(self):
        """
        Gets the message of this ErrorExperiments.


        :return: The message of this ErrorExperiments.
        :rtype: str
        """
        return self._message

    @message.setter
    def message(self, message):
        """
        Sets the message of this ErrorExperiments.


        :param message: The message of this ErrorExperiments.
        :type: str
        """

        self._message = message

    @property
    def more_info(self):
        """
        Gets the more_info of this ErrorExperiments.


        :return: The more_info of this ErrorExperiments.
        :rtype: str
        """
        return self._more_info

    @more_info.setter
    def more_info(self, more_info):
        """
        Sets the more_info of this ErrorExperiments.


        :param more_info: The more_info of this ErrorExperiments.
        :type: str
        """

        self._more_info = more_info

    @property
    def target(self):
        """
        Gets the target of this ErrorExperiments.


        :return: The target of this ErrorExperiments.
        :rtype: ErrorExperimentsTarget
        """
        return self._target

    @target.setter
    def target(self, target):
        """
        Sets the target of this ErrorExperiments.


        :param target: The target of this ErrorExperiments.
        :type: ErrorExperimentsTarget
        """

        self._target = target

    def to_dict(self):
        """
        Returns the model properties as a dict
        """
        result = {}

        for attr, _ in iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """
        Returns the string representation of the model
        """
        return pformat(self.to_dict())

    def __repr__(self):
        """
        For `print` and `pprint`
        """
        return self.to_str()

    def __eq__(self, other):
        """
        Returns true if both objects are equal
        """
        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """
        Returns true if both objects are not equal
        """
        return not self == other



import unittest
import datetime
from sklearn.datasets import load_svmlight_file
from ibm_watson_machine_learning.utils.log_util import get_logger
from preparation_and_cleaning import *
from models_preparation import *


class TestWMLClientWithXGBoost(unittest.TestCase):
    deployment_uid = None
    model_uid = None
    scoring_url = None
    logger = get_logger(__name__)

    @classmethod
    def setUpClass(self):
        TestWMLClientWithXGBoost.logger.info("Service Instance: setting up credentials")

        self.wml_credentials = get_wml_credentials()
        self.client = get_client()
        self.model_path = os.path.join(os.getcwd(), 'artifacts', 'scikit_xgboost_model_' + datetime.now().isoformat())

    def test_1_service_instance_details(self):
        TestWMLClientWithXGBoost.logger.info("Check client ...")
        self.assertTrue(self.client.__class__.__name__ == 'APIClient')

        TestWMLClientWithXGBoost.logger.info("Getting instance details ...")
        details = self.client.service_instance.get_details()

        TestWMLClientWithXGBoost.logger.debug(details)
        self.assertTrue("published_models" in str(details))
        self.assertEqual(type(details), dict)

    def test_2_publish_model(self):
        TestWMLClientWithXGBoost.logger.info("Publishing scikit-xgboost model ...")

        import shutil

        try:
            shutil.rmtree(self.model_path)
        except:
            pass

        create_scikit_learn_xgboost_model_directory(self.model_path)

        self.client.repository.ModelMetaNames.show()

        model_props = {self.client.repository.ModelMetaNames.NAME: "scikit_xg",
            self.client.repository.ModelMetaNames.TYPE: "scikit-learn_0.20",
            self.client.repository.ModelMetaNames.RUNTIME_UID: "scikit-learn_0.20-py3.6"
                       }
        published_model = self.client.repository.store_model(model=self.model_path, meta_props=model_props)
        TestWMLClientWithXGBoost.model_uid = self.client.repository.get_model_uid(published_model)
        TestWMLClientWithXGBoost.logger.info("Published model ID:" + str(TestWMLClientWithXGBoost.model_uid))
        self.assertIsNotNone(TestWMLClientWithXGBoost.model_uid)

        shutil.rmtree(self.model_path)

    def test_3_get_details(self):
        TestWMLClientWithXGBoost.logger.info("Get published model details ...")
        details = self.client.repository.get_details(self.model_uid)

        TestWMLClientWithXGBoost.logger.debug("Model details: " + str(details))
        self.assertTrue("scikit_xg" in str(details))

    def test_4_create_deployment(self):
        TestWMLClientWithXGBoost.logger.info("Create deployment ...")
        global deployment
        deployment = self.client.deployments.create(self.model_uid, meta_props={self.client.deployments.ConfigurationMetaNames.NAME: "Test deployment",self.client.deployments.ConfigurationMetaNames.ONLINE:{}})

        TestWMLClientWithXGBoost.logger.info("model_uid: " + self.model_uid)
        TestWMLClientWithXGBoost.logger.info("Online deployment: " + str(deployment))
        TestWMLClientWithXGBoost.scoring_url = self.client.deployments.get_scoring_href(deployment)
        self.assertTrue("online" in str(deployment))

    def test_5_get_deployment_details(self):
        TestWMLClientWithXGBoost.logger.info("Get deployment details ...")
        deployment_details = self.client.deployments.get_details()
        self.assertTrue("Test deployment" in str(deployment_details))

        TestWMLClientWithXGBoost.logger.debug("Online deployment: " + str(deployment_details))
        TestWMLClientWithXGBoost.deployment_uid = self.client.deployments.get_uid(deployment)
        self.assertIsNotNone(TestWMLClientWithXGBoost.deployment_uid)

    def test_6_score(self):
        TestWMLClientWithXGBoost.logger.info("Online model scoring ...")
        (X, _) = load_svmlight_file(os.path.join('artifacts', 'agaricus.txt.test'))
        scoring_data = {
            self.client.deployments.ScoringMetaNames.INPUT_DATA: [
                {
                    'values': [list(X.toarray()[0, :])]
                }
            ]
        }
        TestWMLClientWithXGBoost.logger.debug("Scoring data: {}".format(scoring_data))
        predictions = self.client.deployments.score(TestWMLClientWithXGBoost.scoring_url, scoring_data)

        TestWMLClientWithXGBoost.logger.debug("Prediction: " + str(predictions))
        self.assertTrue(("prediction" in str(predictions)) and ("probability" in str(predictions)))

    def test_7_delete_deployment(self):
        TestWMLClientWithXGBoost.logger.info("Delete model deployment ...")
        self.client.deployments.delete(TestWMLClientWithXGBoost.deployment_uid)

    def test_8_delete_model(self):
        TestWMLClientWithXGBoost.logger.info("Delete model ...")
        self.client.repository.delete(TestWMLClientWithXGBoost.model_uid)


if __name__ == '__main__':
    unittest.main()



import io
import json
import os
from zipfile import ZipFile
from contextlib import redirect_stdout
from subprocess import check_call
from sys import executable
from tarfile import open as open_tar
from time import sleep
from typing import Dict, Union, Tuple, List, TYPE_CHECKING
from warnings import warn
from functools import wraps

import requests
from ibm_watson_machine_learning.wml_client_error import ApiRequestFailure

from .enums import RunStateTypes
from .errors import (MissingPipeline, AutoAIComputeError, DataFormatNotSupported, LibraryNotCompatible,
                     CannotInstallLibrary, CannotDownloadTrainingDetails, CannotDownloadWMLPipelineDetails,
                     VisualizationFailed)

if TYPE_CHECKING:
    from io import BytesIO, BufferedIOBase
    from pandas import DataFrame
    from collections import OrderedDict
    from sklearn.pipeline import Pipeline
    from ibm_watson_machine_learning import APIClient
    from ibm_watson_machine_learning.helpers import DataConnection, S3Connection
    from ibm_boto3 import resource, client

__all__ = [
    'fetch_pipelines',
    'load_file_from_file_system',
    'NextRunDetailsGenerator',
    'prepare_auto_ai_model_to_publish',
    'remove_file',
    'RunDetailsGenerator',
    'is_ipython',
    'try_import_lale',
    'try_load_dataset',
    'check_dependencies_versions',
    'try_import_autoai_libs',
    'try_import_tqdm',
    'try_import_xlrd',
    'try_import_graphviz',
    'prepare_cos_client',
    'create_model_download_link',
    'create_summary',
    'prepare_auto_ai_model_to_publish_notebook',
    'get_node_and_runtime_index',
    'download_experiment_details_from_file',
    'prepare_model_location_path',
    'download_wml_pipeline_details_from_file',
    'init_cos_client',
    'check_graphviz_binaries'
]


def create_model_download_link(file_path: str):
    """
    Creates download link and shows it in the jupyter notebook

    Parameters
    ----------
    file_path: str, required
    """
    if is_ipython():
        from IPython.display import display
        from ibm_watson_machine_learning.utils import create_download_link
        display(create_download_link(file_path))


def fetch_pipelines(run_params: dict,
                    path: str,
                    pipeline_name: str = None,
                    load_pipelines: bool = False,
                    store: bool = False,
                    wml_client: 'APIClient' = None) -> Union[None, Dict[str, 'Pipeline']]:
    """
    Helper function to download and load computed AutoAI pipelines (sklearn pipelines).

    Parameters
    ----------
    run_params: dict, required
        Fetched details of the run/fit.

    path: str, required
        Local system path indicates where to store downloaded pipelines.

    pipeline_name: str, optional
        Name of the pipeline to download, if not specified, all pipelines are downloaded.

    load_pipelines: bool, optional
        Indicator if we load and return downloaded piepelines.

    store: bool, optional
        Indicator to store pipelines in local filesystem

    wml_client: APIClient, optional
        Should be passed when we are operating on CP4D instance.

    Returns
    -------
    List of sklearn Pipelines or None if load_pipelines is set to False.
    """

    def check_pipeline_nodes(pipeline: dict) -> None:
        """
        Automate check all pipeline nodes to find xgboost or lightgbm dependency.
        """
        # note: check dependencies for estimators
        for node in pipeline['context']['intermediate_model'].get('pipeline_nodes', []):
            if 'XGBClassifierEstimator' == node or 'XGBRegressorEstimator' == node:
                check_dependencies_versions(lib_name='xgboost', error=True)
            elif 'LGBMClassifierEstimator' == node or 'LGBMRegressorEstimator' == node:
                check_dependencies_versions(lib_name='lightgbm', error=True)
        # --- end note

    from sklearn.externals import joblib

    path = os.path.abspath(path)
    pipelines_names = []
    pipelines = {}

    if wml_client is not None:
        model_paths = []

        # note: iterate over all computed pipelines
        for pipeline in run_params['entity']['status'].get('metrics', []):

            # note: fetch and create model paths from file system
            model_path = pipeline['context']['intermediate_model']['location']['model']
            # --- end note

            # note: populate available pipeline names
            if pipeline_name is None:  # checking all pipelines
                model_paths.append(model_path)
                pipelines_names.append(
                    f"Pipeline_{pipeline['context']['intermediate_model']['name'].split('P')[-1]}")

                # note: check dependencies for estimators
                check_pipeline_nodes(pipeline=pipeline)

            # checking only chosen pipeline
            elif pipeline_name == f"Pipeline_{pipeline['context']['intermediate_model']['name'].split('P')[-1]}":
                model_paths.append(model_path)
                pipelines_names = [f"Pipeline_{pipeline['context']['intermediate_model']['name'].split('P')[-1]}"]

                # note: check dependencies for estimators
                check_pipeline_nodes(pipeline=pipeline)

                break
            # --- end note

        if load_pipelines:
            # Disable printing to suppress warning from ai4ml
            with redirect_stdout(open(os.devnull, "w")):
                for model_path, pipeline_name in zip(model_paths, pipelines_names):
                    pipelines[pipeline_name] = joblib.load(load_file_from_file_system(wml_client=wml_client,
                                                                                      file_path=model_path))

        if store:
            for name, pipeline in pipelines.items():
                local_model_path = os.path.join(path, name)
                joblib.dump(pipeline, local_model_path)
                print(f"Selected pipeline stored under: {local_model_path}")

                # note: display download link to the model
                create_model_download_link(local_model_path)
                # --- end note

    else:
        from ibm_boto3 import client
        cos_client = client(
            service_name=run_params['entity']['results_reference']['type'],
            endpoint_url=run_params['entity']['results_reference']['connection']['endpoint_url'],
            aws_access_key_id=run_params['entity']['results_reference']['connection']['access_key_id'],
            aws_secret_access_key=run_params['entity']['results_reference']['connection']['secret_access_key']
        )
        buckets = []
        filenames = []
        keys = []

        for pipeline in run_params['entity']['status'].get('metrics', []):

            if pipeline['context']['phase'] == 'global_output':
                model_path = f"{pipeline['context']['intermediate_model']['location']['model']}"

                if pipeline_name is None:
                    buckets.append(run_params['entity']['results_reference']['location']['bucket'])
                    filenames.append(
                        f"{path}/Pipeline_{pipeline['context']['intermediate_model']['name'].split('P')[-1]}.pickle")
                    keys.append(model_path)
                    pipelines_names.append(
                        f"Pipeline_{pipeline['context']['intermediate_model']['name'].split('P')[-1]}")

                    # note: check dependencies for estimators
                    check_pipeline_nodes(pipeline=pipeline)

                elif pipeline_name == f"Pipeline_{pipeline['context']['intermediate_model']['name'].split('P')[-1]}":
                    buckets = [run_params['entity']['results_reference']['location']['bucket']]
                    filenames = [
                        f"{path}/Pipeline_{pipeline['context']['intermediate_model']['name'].split('P')[-1]}.pickle"]
                    keys = [model_path]
                    pipelines_names = [f"Pipeline_{pipeline['context']['intermediate_model']['name'].split('P')[-1]}"]

                    # note: check dependencies for estimators
                    check_pipeline_nodes(pipeline=pipeline)

                    break

        for bucket, filename, key, name in zip(buckets, filenames, keys, pipelines_names):
            cos_client.download_file(Bucket=bucket, Filename=filename, Key=key)
            if load_pipelines:

                # Disable printing to suppress warning from ai4ml
                with redirect_stdout(open(os.devnull, "w")):
                    pipelines[name] = joblib.load(filename)

                if not store:
                    if os.path.exists(filename):
                        os.remove(filename)

                else:
                    print(f"Selected pipeline stored under: {filename}")

                    # note: display download link to the model
                    create_model_download_link(filename)
                    # --- end note

    if load_pipelines and pipelines:
        return pipelines

    elif load_pipelines:
        raise MissingPipeline(
            pipeline_name if pipeline_name is not None else "global_output pipeline",
            reason="The name of the pipeline is incorrect or there are no pipelines computed.")


def load_file_from_file_system(wml_client: 'APIClient',
                               file_path: str,
                               stream: bool = True) -> 'io.BytesIO':
    """
    Load file into memory from the file system.

    Parameters
    ----------
    wml_client: APIClient, required
        WML v4 client.

    file_path: str, required
        Path in the file system of the file.

    stream: bool, optional
        Indicator to stream data content.

    Returns
    -------
    Sklearn Pipeline
    """
    # note: prepare the file path
    file_path = file_path.split('auto_ml/')[-1]

    if wml_client.default_project_id:
        file_path = f"{file_path}?project_id={wml_client.default_project_id}"

    else:
        file_path = f"{file_path}?space_id={wml_client.default_space_id}"
    # --- end note

    buffer = io.BytesIO()

    response_with_model = requests.get(
        url=f"{wml_client.data_assets._href_definitions.get_wsd_model_attachment_href()}auto_ml/{file_path}",
        headers=wml_client._get_headers(),
        stream=stream,
        verify=False)

    if stream:
        for data in response_with_model.iter_content():
            buffer.write(data)
    else:
        buffer.write(response_with_model.content)

    buffer.seek(0)

    return buffer


class NextRunDetailsGenerator:
    """
    Generator class to produce next list of run details.

    Parameters
    ----------
    wml_client: APIClient, required
        WML Client Instance
    """

    def __init__(self, wml_client: 'APIClient', href: str) -> None:
        self.wml_client = wml_client
        self.next_href = href

    def __iter__(self):
        return self

    def __next__(self):
        if self.next_href is not None:
            response = requests.get(
                url=f"{self.wml_client.wml_credentials['url']}{self.next_href}",
                headers=self.wml_client._get_headers(),
                verify=not self.wml_client.ICP)
            details = response.json()
            self.next_href = details.get('next', {'href': None})['href']
            return details.get('resources', [])

        else:
            raise StopIteration


def prepare_auto_ai_model_to_publish_notebook(pipeline_model: Union['Pipeline', 'TrainablePipeline'],
                                              result_connection,
                                              cos_client) -> Union[Tuple[Dict[str, dict], str], str]:
    """
    Prepares autoai model to publish in Watson Studio via COS.
    Option only for auto-gen notebooks with correct result references on COS.

    Parameters
    ----------
    pipeline_model: Union['Pipeline', 'TrainablePipeline'], required
        model object to publish

    result_connection: DataConnection, required
        Connection object with COS credentials and all needed locations for jsons

    cos_client: ibm_boto3.resource, required
        initialized COS client

    Returns
    -------
    String with path to the saved model and jsons in COS.
    """
    from sklearn.externals import joblib

    artifact_type = ".gzip"

    artifact_name = f"artifact_auto_ai_model{artifact_type}"
    model_artifact_name = f"model_.tar.gz"
    wml_pipeline_definition_name = "pipeline-model.json"
    temp_model_name = '__temp_model.pickle'

    # note: path to the json describing the autoai POD specification
    path = result_connection.location._model_location.split('model.pickle')[0]
    pipeline_model_json_path = f"{path}pipeline-model.json"
    schema_path = f"{path}schema.json"

    bucket = result_connection.location.bucket

    # note: need to download model schema and wml pipeline definition json
    cos_client.meta.client.download_file(Bucket=bucket, Filename=wml_pipeline_definition_name,
                                         Key=pipeline_model_json_path)
    cos_client.meta.client.download_file(Bucket=bucket, Filename='schema.json', Key=schema_path)

    with open('schema.json', 'r') as f:
        schema_json = f.read()

    # note: update the schema, it has wrong field types
    schema_json = schema_json.replace('fieldType', 'type')
    # --- end note

    # note: saved passed model as pickle, for further tar.gz packaging
    joblib.dump(pipeline_model, temp_model_name)
    # --- end note


    # note: create a tar.gz file with model pickle, name it as 'model_run_id.tar.gz', model.pickle inside
    with open_tar(model_artifact_name, 'w:gz') as tar:
        tar.add(temp_model_name, arcname='model.pickle')

    remove_file(filename=temp_model_name)
    # --- end note

    # note: create final zip to publish on WML cloud v4 GA
    with ZipFile(artifact_name, 'w') as zip_file:
        zip_file.write(model_artifact_name)
        zip_file.write(wml_pipeline_definition_name)

    remove_file(filename=model_artifact_name)
    remove_file(filename=wml_pipeline_definition_name)
    # --- end note

    return json.loads(schema_json), artifact_name


def prepare_auto_ai_model_to_publish(
        pipeline_model: Union['Pipeline', 'TrainablePipeline'],
        run_params: dict,
        run_id: str,

        wml_client: 'APIClient') -> Union[Tuple[Dict[str, dict], str], str]:
        """
        Helper function to download and load computed AutoAI pipelines (sklearn pipelines).

        Parameters
        ----------
        pipeline_model: Union['Pipeline', 'TrainablePipeline'], required
            Model that will be prepared for an upload.

        run_params: dict, required
            Fetched details of the run/fit.

        run_id: str, required
            Fit/run ID associated with the model.

        wml_client: APIClient, required

        Returns
        -------
        If cp4d: Dictionary with model schema and artifact name to upload, stored temporally in the user local file system.
        else: path name to the stored model in COS
        """

        from sklearn.externals import joblib

        artifact_type = ".tar.gz" if wml_client.ICP else ".gzip"

        artifact_name = f"artifact_auto_ai_model{artifact_type}"
        model_artifact_name = f"model_{run_id}.tar.gz"
        wml_pipeline_definition_name = "pipeline-model.json"
        temp_model_name = '__temp_model.pickle'

        # note: prepare file paths of pipeline-model and schema (COS / file system location)
        pipeline_info = run_params['entity']['status'].get('metrics')[-1]
        pipeline_model_path = f"{pipeline_info['context']['intermediate_model']['location']['pipeline_model']}"
        schema_path = f"{pipeline_info['context']['intermediate_model']['schema_location']}"
        # --- end note

        if wml_client.ICP:
            # note: downloading pipeline-model.json and schema.json from file system on CP4D
            schema_json = load_file_from_file_system(wml_client=wml_client, file_path=schema_path).read().decode()
            pipeline_model_json = load_file_from_file_system(wml_client=wml_client,
                                                             file_path=pipeline_model_path).read().decode()
            with open(wml_pipeline_definition_name, 'w') as f:
                f.write(pipeline_model_json)
            # --- end note

            # note: update the schema, it has wrong field types
            schema_json = schema_json.replace('fieldType', 'type')
            # --- end note

        else:
            cos_client = init_cos_client(run_params['entity']['results_reference']['connection'])
            bucket = run_params['entity']['results_reference']['location']['bucket']

            # note: need to download model schema and wml pipeline definition json
            cos_client.meta.client.download_file(Bucket=bucket, Filename=wml_pipeline_definition_name,
                                                 Key=pipeline_model_path)
            cos_client.meta.client.download_file(Bucket=bucket, Filename='schema.json', Key=schema_path)

            with open('schema.json', 'r') as f:
                schema_json = f.read()

            # note: update the schema, it has wrong field types
            schema_json = schema_json.replace('fieldType', 'type')
            # --- end note

        # note: saved passed model as pickle, for further tar.gz packaging
        joblib.dump(pipeline_model, temp_model_name)
        # --- end note

        # note: create a tar.gz file with model pickle, name it as 'model_run_id.tar.gz', model.pickle inside
        with open_tar(model_artifact_name, 'w:gz') as tar:
            tar.add(temp_model_name, arcname='model.pickle')

        remove_file(filename=temp_model_name)
        # --- end note

        if wml_client.ICP:
            # note: create final tar.gz to publish on WML (includes pipeline config and tar.gz with model)
            with open_tar(artifact_name, 'w:gz') as tar:
                tar.add(model_artifact_name)
                tar.add(wml_pipeline_definition_name)
        else:
            # note: create final zip to publish on WML cloud v4 GA
            with ZipFile(artifact_name, 'w') as zip_file:
                zip_file.write(model_artifact_name)
                zip_file.write(wml_pipeline_definition_name)

        remove_file(filename=model_artifact_name)
        remove_file(filename=wml_pipeline_definition_name)
        # --- end note

        return json.loads(schema_json), artifact_name


def modify_pipeline_model_json(data_location: str, model_path: str) -> None:
    """
    Change the location of KB model in pipeline-model.json

    Parameters
    ----------
    data_location: str, required
        pipeline-model.json data local path

    model_path: str, required
        Path to KB model stored in COS.
    """
    with open(data_location, 'r') as f:
        data = json.load(f)

    data['pipelines'][0]['nodes'][-1]['parameters']['output_model']['location'] = f"{model_path}model.pickle"

    with open(data_location, 'w') as f:
        f.write(json.dumps(data))


def init_cos_client(connection: dict) -> 'resource':
    """Initiate COS client for further usage."""
    from ibm_botocore.client import Config
    from ibm_boto3 import resource

    if connection.get('auth_endpoint') is not None and connection.get('api_key') is not None:
        cos_client = resource(
            service_name='s3',
            ibm_api_key_id=connection['api_key'],
            ibm_auth_endpoint=connection['auth_endpoint'],
            config=Config(signature_version="oauth"),
            endpoint_url=connection['endpoint_url']
        )

    else:
        cos_client = resource(
            service_name='s3',
            endpoint_url=connection['endpoint_url'],
            aws_access_key_id=connection['access_key_id'],
            aws_secret_access_key=connection['secret_access_key']
        )
    return cos_client


def remove_file(filename: str):
    """Helper function to clean user local storage from temporary package files."""
    if os.path.exists(filename):
        os.remove(filename)


class RunDetailsGenerator:
    """
    Generator class to produce specified run details. Needed for progress bar.

    Parameters
    ----------
    wml_client: APIClient, required
        WML Client Instance

    run_id: str, required
        ID of the run/fit.

    wait_time: float, required
        Waiting time in seconds, how long we will wait until next details will be fetched.
    """

    def __init__(self, wml_client: 'APIClient', run_id: str, wait_time: float) -> None:
        self._wml_client = wml_client
        self.run_id = run_id
        self.wait_time = wait_time
        self.finished = False
        status = self._wml_client.training.get_status(training_uid=self.run_id)
        self.message = status.get('message', {}).get('text', 'Started waiting for resources.').split(': ')[-1]

    def __iter__(self):
        return self

    def __next__(self):
        sleep(self.wait_time)
        try:
            status = self._wml_client.training.get_status(training_uid=self.run_id)
            state = status.get('state')
            message = status.get('message', {}).get('text', 'Started waiting for resources.').split(': ')[-1]

        except ApiRequestFailure as e:
            state = {'state': 'ApiRequestFailure'}
            message = f'{e}, reconnecting...'

        if state == RunStateTypes.FAILED:
            sleep(3)
            try:
                status = self._wml_client.training.get_status(training_uid=self.run_id)
                message = status.get('failure', {}).get('errors', [])
            except ApiRequestFailure as e:
                message = f'{e}, reconnecting...'

            if 'Not Found' in str(message) or 'HeadObject' in str(message):
                raise AutoAIComputeError(self.run_id,
                                         reason=f"Fetching training data error. Please check if COS credentials, "
                                                f"bucket name and path are correct or file system path is correct.")

            elif 'Failed_Image' in str(message):
                raise AutoAIComputeError(
                    self.run_id,
                    reason=f"This version of WML V4 SDK with AutoAI component is no longer supported. "
                           f"Please update your package to the newest version. "
                           f"Error: {message}")

            else:
                raise AutoAIComputeError(self.run_id, reason=f"Error: {message}")

        # note: additional checks, OBM produces similar completed message state, we need to finish only when KB finished
        elif state == RunStateTypes.COMPLETED and 'Training job' in message and 'completed' in message:
            if self.finished:
                raise StopIteration
            else:
                self.finished = True
                return message

        else:
            return message


def is_ipython():
    """Check if code is running in the notebook."""
    try:
        name = get_ipython().__class__.__name__
        if name != 'ZMQInteractiveShell':
            return False
        else:
            return True

    except Exception:
        return False


def try_import_lale():
    """
    Check if lale package is installed in local environment, if not, just download and install it.
    """
    lale_version = '0.3.18'
    try:
        import lale
        from packaging import version
        if version.parse(lale.__version__) < version.parse(lale_version):
            warn(f"\"lale\" package version is to low."
                 f"Installing version >={lale_version}")

            try:
                check_call([executable, "-m", "pip", "install", f"lale>={lale_version}"])

            except Exception as e:
                raise CannotInstallLibrary(value_name=e,
                                           reason="lale failed to install. Please install it manually.")

    except AttributeError:
        warn(f"Cannot determine \"lale\" package version."
             f"Installing version >={lale_version}")

        try:
            check_call([executable, "-m", "pip", "install", f"lale>={lale_version}"])

        except Exception as e:
            raise CannotInstallLibrary(value_name=e,
                                       reason="lale failed to install. Please install it manually.")

    except ImportError:
        warn(f"\"lale\" package is not installed. "
             f"This is the needed dependency for pipeline model refinery, we will try to install it now...")

        try:
            check_call([executable, "-m", "pip", "install", f"lale>={lale_version}"])

        except Exception as e:
            raise CannotInstallLibrary(value_name=e,
                                       reason="lale failed to install. Please install it manually.")


def try_import_autoai_libs():
    """
    Check if autoai_libs package is installed in local environment, if not, just download and install it.
    """
    try:
        import autoai_libs

    except ImportError:
        warn(f"\"autoai_libs\" package is not installed. "
             f"This is the needed dependency for pipeline model refinery, we will try to install it now...")

        try:
            check_call([executable, "-m", "pip", "install", "autoai_libs>=1.11.0"])

        except Exception as e:
            raise CannotInstallLibrary(value_name=e,
                                       reason="autoai_libs>=1.11.0 failed to install. Please install it manually.")


def try_import_tqdm():
    """
    Check if tqdm package is installed in local environment, if not, just download and install it.
    """
    try:
        import tqdm

    except ImportError:
        warn(f"\"tqdm\" package is not installed. "
             f"This is the needed dependency for pipeline training, we will try to install it now...")

        try:
            check_call([executable, "-m", "pip", "install", "tqdm==4.43.0"])

        except Exception as e:
            raise CannotInstallLibrary(value_name=e,
                                       reason="tqdm==4.43.0 failed to install. Please install it manually.")


def try_import_xlrd():
    """
    Check if xlrd package is installed in local environment, if not, just download and install it.
    """
    try:
        import xlrd

    except ImportError:
        warn(f"\"xlrd\" package is not installed. "
             f"This is the needed dependency for loading dataset from xls files, we will try to install it now...")

        try:
            check_call([executable, "-m", "pip", "install", "xlrd==1.2.0"])

        except Exception as e:
            raise CannotInstallLibrary(value_name=e,
                                       reason="xlrd==1.2.0 failed to install. Please install it manually.")


def try_import_graphviz():
    """
    Check if graphviz package is installed in local environment, if not, just download and install it.
    """
    try:
        import graphviz

    except ImportError:
        warn(f"\"graphviz\" package is not installed. "
             f"This is the needed dependency for visualizing data join graph, we will try to install it now...")

        try:
            check_call([executable, "-m", "pip", "install", "graphviz==0.14"])

        except Exception as e:
            raise CannotInstallLibrary(value_name=e,
                                       reason="graphviz==0.14 failed to install. Please install it manually.")


def try_load_dataset(buffer: Union['BytesIO', 'BufferedIOBase'], sheet_name: str = 0, separator: str = ',') -> Union[
    'DataFrame', 'OrderedDict']:
    """
    Load data into a pandas DataFrame from BytesIO object.

    Parameters
    ----------
    buffer: Union['BytesIO', 'BufferedIOBase'], required
        Buffer with bytes data.

    sheet_name: str, optional
        Name of the xlsx sheet to read.

    separator: str, optional
        csv separator

    Returns
    -------
    DataFrame or OrderedDict
    """
    from pandas import read_csv, read_excel

    try:
        buffer.seek(0)
        data = read_csv(buffer, sep=separator, encoding='utf-8')

    except Exception as e1:
        try:
            try_import_xlrd()
            buffer.seek(0)
            data = read_excel(buffer, sheet_name=sheet_name)

        except Exception as e2:
            raise DataFormatNotSupported(None, reason=f"Error1: {e1} Error2: {e2}")

    return data


def check_dependencies_versions(lib_name: str = None, error: bool = False, msg: str = '') -> None:
    """Check scikit, xgboost and lightbm installed versions and inform the user about needed ones.

    Parameters
    ----------
    lib_name: str, optional
        Library name to check

    error: bool, optional
        If True, raise an error when some library is missing or is incompatible.

    msg: str, optional
        Error message.
    """

    scikit_version = '0.20.3'
    numpy_version = '1.18.2'
    xgboost_version = '0.90'
    lightgbm_version = '2.2.3'

    scikit_message = f'To use AutoAI, you need to have scikit-learn {scikit_version} installed.'
    numpy_message = f'To use AutoAI, you need to have numpy {numpy_version} installed.'
    xgboost_message = f'To use AutoAI with xgboost estimators, you need to have xgboost {xgboost_version} installed.'
    lightgbm_message = f'To use AutoAI with lightgbm estimators, you need to have lightgbm {lightgbm_version} installed.'

    def print_or_warn(library: str) -> None:
        if library == 'scikit':
            message = scikit_message

        elif library == 'xgboost':
            message = xgboost_message

        elif library == 'numpy':
            message = numpy_message

        else:
            message = lightgbm_message

        if error:
            raise LibraryNotCompatible(reason=f'{msg} {message}')

        if is_ipython():
            print(f'Warning: {message}')

        else:
            warn(message=message)

    # note: scikit version check
    try:
        if lib_name == 'scikit' or lib_name is None:
            import sklearn
            if sklearn.__version__ != scikit_version:
                print_or_warn(library='scikit')

    except (ImportError, AttributeError, ModuleNotFoundError):
        print_or_warn(library='scikit')
    # --- end note

    # note: xgboost version check
    try:
        if lib_name == 'xgboost' or lib_name is None:
            import xgboost
            if xgboost.__version__ != xgboost_version:
                print_or_warn(library='xgboost')

    except (ImportError, AttributeError, ModuleNotFoundError):
        print_or_warn(library='xgboost')
    # --- end note

    # note: lightgbm version check
    try:
        if lib_name == 'lightgbm' or lib_name is None:
            import lightgbm
            if lightgbm.__version__ != lightgbm_version:
                print_or_warn(library='lightgbm')

    except (ImportError, AttributeError, ModuleNotFoundError):
        print_or_warn(library='lightgbm')
    # --- end note

    # note: numpy version check
    try:
        if lib_name == 'numpy' or lib_name is None:
            import numpy
            from packaging import version
            if version.parse(numpy.__version__) < version.parse(numpy_version):
                print_or_warn(library='numpy')

    except (ImportError, AttributeError, ModuleNotFoundError):
        print_or_warn(library='numpy')
    # --- end note


def prepare_cos_client(
        training_data_references: List['DataConnection'] = None,
        training_result_reference: 'DataConnection' = None) -> Tuple[Union[List[Tuple['DataConnection', 'resource']]],
                                                                     Union[Tuple['DataConnection', 'resource'], None]]:
    """
    Create COS clients for training data and results.

    Parameters
    ----------
    training_data_references: List['DataConnection'], optional

    training_result_reference: 'DataConnection', optional

    Returns
    -------
    list of COS clients for training data , client for results
    """
    from ibm_watson_machine_learning.helpers import S3Connection
    from ibm_boto3 import resource
    from ibm_botocore.client import Config

    def differentiate_between_credentials(connection: 'S3Connection') -> 'resource':
        # note: we do not know which version of COS credentials user used during training
        if hasattr(connection, 'auth_endpoint') and hasattr(connection, 'api_key'):
            cos_client = resource(
                service_name='s3',
                ibm_api_key_id=connection.api_key,
                ibm_auth_endpoint=connection.auth_endpoint,
                config=Config(signature_version="oauth"),
                endpoint_url=connection.endpoint_url
            )

        else:
            cos_client = resource(
                service_name='s3',
                endpoint_url=connection.endpoint_url,
                aws_access_key_id=connection.access_key_id,
                aws_secret_access_key=connection.secret_access_key
            )
        # --- end note

        return cos_client

    cos_client_results = None
    data_cos_clients = []

    if training_result_reference is not None:
        if isinstance(training_result_reference.connection, S3Connection):
            cos_client_results = (training_result_reference,
                                  differentiate_between_credentials(connection=training_result_reference.connection))

    if training_data_references is not None:
        for reference in training_data_references:
            if isinstance(reference.connection, S3Connection):
                data_cos_clients.append((reference,
                                         differentiate_between_credentials(connection=reference.connection)))

    return data_cos_clients, cos_client_results


def create_summary(details: dict, scoring: str) -> 'DataFrame':
    """
    Creates summary in a form of a pandas.DataFrame of computed pipelines (should be used in remote and local scenario
    with COS).

    Parameters
    ----------
    details: dict, required
        Dictionary with all training data

    scoring: str, required
        scoring method

    Returns
    -------
    pandas.DataFrame with pipelines summary
    """
    from ibm_watson_machine_learning.utils.autoai.enums import MetricsToDirections, Directions
    from pandas import DataFrame

    columns = (['Pipeline Name', 'Number of enhancements', 'Estimator'] +
               [metric_name for metric_name in
                details['entity']['status'].get('metrics', [{}])[0].get('ml_metrics', {}).keys()])
    values = []

    for pipeline in details['entity']['status'].get('metrics', []):
        if pipeline['context']['phase'] == 'global_output':
            number_of_enhancements = len(pipeline['context']['intermediate_model']['composition_steps']) - 5

            # note: workaround when some pipelines have less or more metrics computed
            metrics = columns[3:]
            pipeline_metrics = [None] * len(metrics)
            for metric, value in pipeline['ml_metrics'].items():
                for i, metric_name in enumerate(metrics):
                    if metric_name == metric:
                        pipeline_metrics[i] = value
            # --- end note

            values.append(
                ([f"Pipeline_{pipeline['context']['intermediate_model']['name'].split('P')[-1]}"] +
                 [number_of_enhancements] +
                 [pipeline['context']['intermediate_model']['pipeline_nodes'][-1]] +
                 pipeline_metrics
                 ))

    pipelines = DataFrame(data=values, columns=columns)
    pipelines.drop_duplicates(subset="Pipeline Name", keep='first', inplace=True)
    pipelines.set_index('Pipeline Name', inplace=True)

    try:
        if (MetricsToDirections[scoring.upper()].value ==
                Directions.ASCENDING):
            return pipelines.sort_values(
                by=[f"training_{scoring}"], ascending=False).rename(
                {
                    f"training_{scoring}":
                        f"training_{scoring}_(optimized)"
                }, axis='columns')

        else:
            return pipelines.sort_values(by=[f"training_{scoring}"]).rename(
                {
                    f"training_{scoring}":
                        f"training_{scoring}_(optimized)"
                }, axis='columns')

    # note: sometimes backend will not return 'training_' prefix to the metric
    except KeyError:
        return pipelines


def get_node_and_runtime_index(node_name: str, optimizer_config: dict) -> Tuple[int, int]:
    """Find node index from node name in experiment parameters."""
    node_number = None
    runtime_number = None

    for i, node in enumerate(optimizer_config['entity']['document']['pipelines'][0]['nodes']):
        if node_name == 'kb' and (node.get('id') == 'kb' or node.get('id') == 'automl'):
            node_number = i
            break

        elif node_name == 'obm' and node.get('id') == 'obm':
            node_number = i
            break

    for i, runtime in enumerate(optimizer_config['entity']['document']['runtimes']):
        if node_name == 'kb' and (runtime.get('id') == 'kb' or runtime.get('id') == 'automl' or
                                  runtime.get('id') == 'autoai'):
            runtime_number = i
            break

        elif node_name == 'obm' and runtime.get('id') == 'obm':
            runtime_number = i
            break

    return node_number, runtime_number


def download_experiment_details_from_file(result_client_and_connection: Tuple['DataConnection', 'resource']) -> dict:
    """Try to download training details from user COS."""

    try:
        file = result_client_and_connection[1].Object(
            result_client_and_connection[0].location.bucket,
            result_client_and_connection[0].location._training_status).get()

        details = json.loads(file['Body'].read())

    except Exception as e:
        raise CannotDownloadTrainingDetails('', reason=f"Error: {e}")

    return details


def download_wml_pipeline_details_from_file(result_client_and_connection: Tuple['DataConnection', 'resource']) -> dict:
    """Try to download wml pipeline details from user COS."""

    try:
        path = result_client_and_connection[0].location._model_location.split('model.pickle')[0]
        path = f"{path}pipeline-model.json"

        file = result_client_and_connection[1].Object(
            result_client_and_connection[0].location.bucket,
            path).get()

        details = json.loads(file['Body'].read())

    except Exception as e:
        raise CannotDownloadWMLPipelineDetails('', reason=f"Error: {e}")

    return details


def prepare_model_location_path(model_path: str) -> str:
    """
    To be able to get best pipeline after computation we need to change model_location string to global_output.
    """

    if "data/automl/" in model_path:
        path = model_path.split('data/automl/')[0]
        path = f"{path}data/automl/global_output/"

    else:
        path = model_path.split('data/kb/')[0]
        path = f"{path}data/kb/global_output/"

    return path


def check_graphviz_binaries(f):
    @wraps(f)
    def _f(*method_args, **method_kwargs):
        from graphviz.backend import ExecutableNotFound
        try:
            output = f(*method_args, **method_kwargs)

        except ExecutableNotFound as e:
            raise VisualizationFailed(
                reason=f"Cannot perform visualization with graphviz. Please make sure that you have Graphviz binaries "
                       f"installed in your system. Please follow this guide: https://www.graphviz.org/download/")

        return output

    return _f



import unittest
import sys
import io
from multiprocessing import Process, Queue
from ibm_watson_machine_learning.utils.log_util import get_logger
from ibm_watson_machine_learning.experiments import Experiments
from preparation_and_cleaning import *


class TestWMLClientWithExperiment(unittest.TestCase):
    deployment_uid = None
    model_uid = None
    model2_uid = None
    scoring_url = None
    definition_1_uid = None
    definition_1_url = None
    trained_model_uid = None
    experiment_uid = None
    experiment_run_uid = None
    trained_model_uids = None
    logger = get_logger(__name__)

    @classmethod
    def setUpClass(self):
        TestWMLClientWithExperiment.logger.info("Service Instance: setting up credentials")
        self.wml_credentials = get_wml_credentials()
        # reload(site)
        self.client = get_client()
        self.cos_resource = get_cos_resource()
        self.bucket_names = prepare_cos(self.cos_resource)

    @classmethod
    def tearDownClass(self):
        clean_cos(self.cos_resource, self.bucket_names)
        pass

    def test_01_service_instance_details(self):
        TestWMLClientWithExperiment.logger.info("Check client ...")
        self.assertTrue(self.client.__class__.__name__ == 'APIClient')

        TestWMLClientWithExperiment.logger.info("Getting instance details ...")
        details = self.client.service_instance.get_details()
        TestWMLClientWithExperiment.logger.debug(details)

        self.assertTrue("published_models" in str(details))
        self.assertEqual(type(details), dict)

    def test_02_save_definition1(self):
        import os

        TestWMLClientWithExperiment.logger.info("Save model definition ...")
        metadata = {
            self.client.repository.DefinitionMetaNames.NAME: "DDL with metrics",
            self.client.repository.DefinitionMetaNames.DESCRIPTION: "my_description",
            self.client.repository.DefinitionMetaNames.AUTHOR_NAME: "John Smith",
            self.client.repository.DefinitionMetaNames.AUTHOR_EMAIL: "js@js.com",
            self.client.repository.DefinitionMetaNames.FRAMEWORK_NAME: "tensorflow-ddl",
            self.client.repository.DefinitionMetaNames.FRAMEWORK_VERSION: "1.5",
            self.client.repository.DefinitionMetaNames.RUNTIME_NAME: "python",
            self.client.repository.DefinitionMetaNames.RUNTIME_VERSION: "3.5",
            self.client.repository.DefinitionMetaNames.EXECUTION_COMMAND: "python mnist-env.py"
            }

        model_content_path = os.path.join('artifacts', 'tf_distributed_ddl', 'tf_ddl.zip')
        definition_details = self.client.repository.store_definition(training_definition=model_content_path, meta_props=metadata)
        TestWMLClientWithExperiment.definition_1_uid = self.client.repository.get_definition_uid(definition_details)
        TestWMLClientWithExperiment.definition_1_url = self.client.repository.get_definition_url(definition_details)
        TestWMLClientWithExperiment.logger.info("Saved model definition uid: " + str(TestWMLClientWithExperiment.definition_1_uid))

    def test_03_get_definition_details(self):
        TestWMLClientWithExperiment.logger.info("Getting definition details ...")
        details_1 = self.client.repository.get_definition_details(TestWMLClientWithExperiment.definition_1_uid)
        TestWMLClientWithExperiment.logger.info(details_1)
        self.assertTrue('DDL with metrics' in str(details_1))

    def test_04_save_experiment(self):
        metadata = {
                    self.client.repository.ExperimentMetaNames.NAME: "tf_distributed_DDL_experiment",
                    self.client.repository.ExperimentMetaNames.DESCRIPTION: "mnist best model",
                    self.client.repository.ExperimentMetaNames.AUTHOR_NAME: "John Smith",
                    self.client.repository.ExperimentMetaNames.AUTHOR_EMAIL: "js@js.com",
                    self.client.repository.ExperimentMetaNames.TRAINING_DATA_REFERENCE: get_cos_training_data_reference(self.bucket_names),
                    self.client.repository.ExperimentMetaNames.TRAINING_RESULTS_REFERENCE: get_cos_training_results_reference(self.bucket_names),
                    self.client.repository.ExperimentMetaNames.TRAINING_REFERENCES: [
                        {
                            "name": "mnist_DDL_distributed",
                            "training_definition_url": TestWMLClientWithExperiment.definition_1_url,
                            "compute_configuration": {"name": "p100x2", "nodes": 1},
                        }
                    ]
                }

        experiment_details = self.client.repository.store_experiment(meta_props=metadata)
        self.assertTrue('tf_distributed_DDL_experiment' in str(experiment_details))

        TestWMLClientWithExperiment.experiment_uid = self.client.repository.get_experiment_uid(experiment_details)

        experiment_specific_details = self.client.repository.get_experiment_details(TestWMLClientWithExperiment.experiment_uid)
        self.assertTrue(TestWMLClientWithExperiment.experiment_uid in str(experiment_specific_details))

    def test_06_get_experiment_details(self):
        details = self.client.repository.get_experiment_details()
        self.assertTrue(TestWMLClientWithExperiment.experiment_uid in str(details))

        details2 = self.client.repository.get_experiment_details(TestWMLClientWithExperiment.experiment_uid)
        self.assertTrue(TestWMLClientWithExperiment.experiment_uid in str(details2))

    def test_07_run_experiment(self):
        created_experiment_run_details = self.client.experiments.run(TestWMLClientWithExperiment.experiment_uid, asynchronous=False)
        self.assertIsNotNone(created_experiment_run_details)
        TestWMLClientWithExperiment.experiment_run_uid = Experiments.get_run_uid(created_experiment_run_details)

    def test_08_get_status(self):
        status = self.client.experiments.get_status(TestWMLClientWithExperiment.experiment_run_uid)
        self.assertIsNotNone(status)
        self.assertTrue(status['state'] == 'completed')

    def test_09_monitor_logs(self):
        training_uids = self.client.experiments.get_training_uids(self.client.experiments.get_run_details(TestWMLClientWithExperiment.experiment_run_uid))

        queue = Queue()
        process = Process(target=run_monitor_training, args=(self.client, training_uids[0], queue))
        process.start()
        process.join(timeout=90)
        process.terminate()
        self.assertFalse(queue.empty())
        self.assertTrue("GPU" in queue.get())

    def test_10_get_all_experiments_runs_details(self):
        details = self.client.experiments.get_details()
        self.assertIsNotNone(details)

    def test_11_get_experiment_details(self):
        details = self.client.experiments.get_details(TestWMLClientWithExperiment.experiment_uid)
        self.assertIsNotNone(details)

    def test_12_get_experiment_run_details(self):
        details = self.client.experiments.get_run_details(TestWMLClientWithExperiment.experiment_run_uid)
        self.assertIsNotNone(details)

    def test_13_list_experiments(self):
        stdout_ = sys.stdout
        captured_output = io.StringIO()  # Create StringIO object
        sys.stdout = captured_output  # and redirect stdout.
        self.client.experiments.list_runs()  # Call function.
        sys.stdout = stdout_  # Reset redirect.
        self.assertTrue(TestWMLClientWithExperiment.experiment_uid in captured_output.getvalue())
        self.client.experiments.list_runs()  # Just to see values.

    def test_14_list_experiment_runs_for_experiment(self):
        stdout_ = sys.stdout
        captured_output = io.StringIO()  # Create StringIO object
        sys.stdout = captured_output  # and redirect stdout.
        self.client.experiments.list_runs(TestWMLClientWithExperiment.experiment_uid)  # Call function.
        sys.stdout = stdout_  # Reset redirect.
        self.assertTrue(TestWMLClientWithExperiment.experiment_uid in captured_output.getvalue())
        self.client.experiments.list_runs(TestWMLClientWithExperiment.experiment_uid)  # Just to see values.

    def test_15_list_training_runs_for_experiment_run(self):
        stdout_ = sys.stdout
        captured_output = io.StringIO()  # Create StringIO object
        sys.stdout = captured_output  # and redirect stdout.
        self.client.experiments.list_training_runs(TestWMLClientWithExperiment.experiment_run_uid)  # Call function.
        sys.stdout = stdout_  # Reset redirect.
        self.assertTrue('GUID' in captured_output.getvalue())
        self.assertTrue('completed' in captured_output.getvalue())
        self.client.experiments.list_training_runs(TestWMLClientWithExperiment.experiment_run_uid)  # to see values

    def test_16_get_training_runs_uid(self):
        details = self.client.experiments.get_run_details(TestWMLClientWithExperiment.experiment_run_uid)
        TestWMLClientWithExperiment.trained_model_uids = self.client.experiments.get_training_uids(details)
        self.assertTrue(len(TestWMLClientWithExperiment.trained_model_uids) == 1)

    """
    def test_17_experiment_metrics(self):
        print("Get experiment metrics ..")
        metrics = self.client.experiments.get_metrics(TestWMLClientWithExperiment.experiment_run_uid)
        final_metrics = self.client.experiments.get_latest_metrics(TestWMLClientWithExperiment.experiment_run_uid)

        print('metrics: ' + str(metrics))
        print('final metrics: ' + str(final_metrics))

        self.assertTrue('accuracy' in str(metrics))
        self.assertTrue('accuracy' in str(final_metrics))
        self.assertTrue(len(final_metrics) == 1)

    def test_18_store_trained_models(self):
        model_1 = self.client.repository.store_model(TestWMLClientWithExperiment.trained_model_uids[0], {'name': 'distributed mnist'})
        TestWMLClientWithExperiment.model_uid = self.client.repository.get_model_uid(model_1)

        print(TestWMLClientWithExperiment.model_uid)
        self.assertIsNotNone(TestWMLClientWithExperiment.model_uid)
    """
    def test_19_delete_experiment_run(self):
        self.client.experiments.delete(TestWMLClientWithExperiment.experiment_run_uid)

    def test_20_list_repository_1(self):
        captured_output = io.StringIO()  # Create StringIO object
        sys.stdout = captured_output  # and redirect stdout.
        self.client.repository.list()  # Call function.
        sys.stdout = sys.__stdout__  # Reset redirect.
        self.assertTrue(TestWMLClientWithExperiment.definition_1_uid in captured_output.getvalue())
        self.assertTrue(TestWMLClientWithExperiment.experiment_uid in captured_output.getvalue())
        self.client.repository.list()  # Just to see values.

    def test_21_delete_experiment(self):
        self.client.repository.delete(TestWMLClientWithExperiment.experiment_uid)

    def test_22_delete_definitions(self):
        self.client.repository.delete(TestWMLClientWithExperiment.definition_1_uid)

    def test_23_list_repository_2(self):
        stdout_ = sys.stdout
        captured_output = io.StringIO()  # Create StringIO object
        sys.stdout = captured_output  # and redirect stdout.
        self.client.repository.list()  # Call function.
        sys.stdout = stdout_  # Reset redirect.
        self.assertTrue(TestWMLClientWithExperiment.definition_1_uid not in captured_output.getvalue())
        self.assertTrue(TestWMLClientWithExperiment.experiment_uid not in captured_output.getvalue())
        self.client.repository.list()  # Just to see values.

    """
    def test_24_get_models_details(self):
        print(TestWMLClientWithExperiment.model_uid)
        self.assertTrue(TestWMLClientWithExperiment.model_uid in str(self.client.repository.get_model_details(TestWMLClientWithExperiment.model_uid)))

    def test_25_delete_models(self):
        self.client.repository.delete(TestWMLClientWithExperiment.model_uid)
        self.client.repository.delete(TestWMLClientWithExperiment.model2_uid)
    """


if __name__ == '__main__':
    unittest.main()



import unittest
from ibm_watson_machine_learning.utils.log_util import get_logger
from ibm_watson_machine_learning.wml_client_error import ApiRequestFailure
from ibm_watson_machine_learning.tests.Cloud.preparation_and_cleaning import *


class TestWMLClientWithPMML(unittest.TestCase):
    deployment_uid = None
    model_uid = None
    scoring_url = None
    scoring_uid = None
    logger = get_logger(__name__)

    @classmethod
    def setUpClass(self):
        TestWMLClientWithPMML.logger.info("Service Instance: setting up credentials")

        self.wml_credentials = get_wml_credentials()
        self.client = get_client()

        self.cos_credentials = get_cos_credentials()
        self.cos_resource_crn = self.cos_credentials['resource_instance_id']
        self.instance_crn = get_instance_crn()

        self.space_name = str(uuid.uuid4())

        metadata = {
            self.client.spaces.ConfigurationMetaNames.NAME: 'space' + self.space_name,
            self.client.spaces.ConfigurationMetaNames.DESCRIPTION: self.space_name + ' description',
            self.client.spaces.ConfigurationMetaNames.STORAGE: {
                "type": "bmcos_object_storage",
                "resource_crn": self.cos_resource_crn
            },
            self.client.spaces.ConfigurationMetaNames.COMPUTE: {
                         "name": "test_pmml",
                         "crn": self.instance_crn
            }
        }

        self.space = self.client.spaces.store(meta_props=metadata)

        TestWMLClientWithPMML.space_id = self.client.spaces.get_id(self.space)
        print("space_id: ", TestWMLClientWithPMML.space_id)
        self.client.set.default_space(TestWMLClientWithPMML.space_id)

    def test_01_publish_local_model_in_repository(self):

        self.client.repository.ModelMetaNames.show()

        model_content_path = 'artifacts/pmml-model.xml'

        sw_spec_uid = self.client.software_specifications.get_uid_by_name("spark-mllib_2.4")

        meta_props = {self.client.repository.ModelMetaNames.NAME: "test_pmml",
                      self.client.repository.ModelMetaNames.SOFTWARE_SPEC_UID: sw_spec_uid,
                      self.client.repository.ModelMetaNames.TYPE: 'pmml_1.1'}

        model_details = self.client.repository.store_model(model=model_content_path, meta_props=meta_props)

        TestWMLClientWithPMML.model_id = self.client.repository.get_model_uid(model_details)

        print(TestWMLClientWithPMML.model_id)

        self.assertIsNotNone(TestWMLClientWithPMML.model_id)

    def test_02_get_details(self):
        TestWMLClientWithPMML.logger.info("Get details")
        self.assertIsNotNone(self.client.repository.get_details(TestWMLClientWithPMML.model_id))

    def test_03_create_deployment(self):
        TestWMLClientWithPMML.logger.info("Create deployment")
        deployment_details = self.client.deployments.create(TestWMLClientWithPMML.model_id,
                                                            meta_props={self.client.deployments.ConfigurationMetaNames.NAME: "Test pmml deployment",
                                                                        self.client.deployments.ConfigurationMetaNames.ONLINE:{}})

        TestWMLClientWithPMML.logger.debug("Deployment details: {}".format(deployment_details))

        TestWMLClientWithPMML.deployment_id = self.client.deployments.get_id(deployment_details)

        TestWMLClientWithPMML.scoring_url = self.client.deployments.get_scoring_href(deployment_details)
        TestWMLClientWithPMML.logger.debug("Scoring url: {}".format(TestWMLClientWithPMML.scoring_url))

        self.assertTrue('online' in str(deployment_details))

    def test_04_get_deployment_details(self):
        TestWMLClientWithPMML.logger.info("Get deployment details")
        deployment_details = self.client.deployments.get_details(deployment_uid=TestWMLClientWithPMML.deployment_id)
        self.assertIsNotNone(deployment_details)

    def test_05_delete_model(self):
        TestWMLClientWithPMML.logger.info("Delete model")
        self.client.deployments.delete(TestWMLClientWithPMML.deployment_id)
        self.client.repository.delete(TestWMLClientWithPMML.model_id)

    def test_06_delete_space(self):
        TestWMLClientWithPMML.logger.info("Delete space")
        self.client.spaces.delete(TestWMLClientWithPMML.space_id)

if __name__ == '__main__':
    unittest.main()
